# Copyright 2017 - Nokia Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import mock
from six import BytesIO

from glare.common import exception
from glare.common import store_api
from glare.tests.unit import base


class TestStaticQuotas(base.BaseTestArtifactAPI):
    """Test static quota limits."""

    def test_count_artifact_number(self):
        user1_req = self.get_fake_request(self.users['user1'])
        user2_req = self.get_fake_request(self.users['user2'])
        # initially there are no artifacts
        self.assertEqual(
            0, len(self.controller.list(user1_req, 'all')['artifacts']))
        self.assertEqual(
            0, len(self.controller.list(user2_req, 'all')['artifacts']))

        # set global limit on 10 artifacts
        self.config(max_artifact_number=10)
        # 3 images, 15 heat templates, 10 murano packages
        self.config(max_artifact_number=3,
                    group='artifact_type:images')
        self.config(max_artifact_number=15,
                    group='artifact_type:heat_templates')
        self.config(max_artifact_number=10,
                    group='artifact_type:murano_packages')

        # create 3 images for user1
        for i in range(3):
            img = self.controller.create(
                user1_req, 'images', {'name': 'img%d' % i})

        # creation of another image fails because of artifact type limit
        self.assertRaises(exception.Forbidden, self.controller.create,
                          user1_req, 'images', {'name': 'img4'})

        # create 7 murano packages
        for i in range(7):
            self.controller.create(
                user1_req, 'murano_packages', {'name': 'mp%d' % i})

        # creation of another package fails because of global limit
        self.assertRaises(exception.Forbidden, self.controller.create,
                          user1_req, 'murano_packages', {'name': 'mp8'})

        # delete an image and create another murano package work
        self.controller.delete(user1_req, 'images', img['id'])
        self.controller.create(user1_req, 'murano_packages', {'name': 'mp8'})

        # user2 can create his own artifacts
        for i in range(10):
            self.controller.create(
                user2_req, 'heat_templates', {'name': 'ht%d' % i})

        # creation of another heat template fails because of global limit
        self.assertRaises(exception.Forbidden, self.controller.create,
                          user2_req, 'heat_templates', {'name': 'ht11'})

        # disable global limit and try to create 15 heat templates
        self.config(max_artifact_number=-1)
        for i in range(15):
            self.controller.create(
                user1_req, 'heat_templates', {'name': 'ht%d' % i})

        # creation of another heat template fails because of type limit
        self.assertRaises(exception.Forbidden, self.controller.create,
                          user1_req, 'heat_templates', {'name': 'ht16'})

        # disable type limit for heat templates and create 1 heat templates
        self.config(max_artifact_number=-1,
                    group='artifact_type:heat_templates')
        self.controller.create(
            user1_req, 'heat_templates', {'name': 'ht16'})

    def test_calculate_uploaded_data(self):
        user1_req = self.get_fake_request(self.users['user1'])
        user2_req = self.get_fake_request(self.users['user2'])
        # initially there are no artifacts
        self.assertEqual(
            0, len(self.controller.list(user1_req, 'all')['artifacts']))
        self.assertEqual(
            0, len(self.controller.list(user2_req, 'all')['artifacts']))

        # set global limit on 1000 bytes
        self.config(max_uploaded_data=1000)
        # 300 for sample artifact, 1500 for images, 1000 for murano packages
        self.config(max_uploaded_data=300,
                    group='artifact_type:sample_artifact')
        self.config(max_uploaded_data=1500,
                    group='artifact_type:images')
        self.config(max_uploaded_data=1000,
                    group='artifact_type:murano_packages')

        # create 2 sample artifacts for user 1
        art1 = self.controller.create(
            user1_req, 'sample_artifact', {'name': 'art1'})
        art2 = self.controller.create(
            user1_req, 'sample_artifact', {'name': 'art2'})

        # create 3 images for user1
        img1 = self.controller.create(
            user1_req, 'images', {'name': 'img1'})
        img2 = self.controller.create(
            user1_req, 'images', {'name': 'img2'})
        img3 = self.controller.create(
            user1_req, 'images', {'name': 'img3'})

        # upload to art1 fails now because of type limit
        self.assertRaises(
            exception.RequestEntityTooLarge, self.controller.upload_blob,
            user1_req, 'sample_artifact', art1['id'], 'blob',
            BytesIO(b'a' * 301), 'application/octet-stream', 301)

        # upload to img1 fails now because of global limit
        self.assertRaises(
            exception.RequestEntityTooLarge, self.controller.upload_blob,
            user1_req, 'images', img1['id'], 'image',
            BytesIO(b'a' * 1001), 'application/octet-stream', 1001)

        # upload 300 bytes to 'blob' of art1
        self.controller.upload_blob(
            user1_req, 'sample_artifact', art1['id'], 'blob',
            BytesIO(b'a' * 300), 'application/octet-stream',
            content_length=300)

        # upload another blob to art1 fails because of type limit
        self.assertRaises(
            exception.RequestEntityTooLarge, self.controller.upload_blob,
            user1_req, 'sample_artifact', art1['id'],
            'dict_of_blobs/blob', BytesIO(b'a'),
            'application/octet-stream', 1)

        # upload to art2 fails now because of type limit
        self.assertRaises(
            exception.RequestEntityTooLarge, self.controller.upload_blob,
            user1_req, 'sample_artifact', art2['id'], 'blob',
            BytesIO(b'a'), 'application/octet-stream', 1)

        # delete art1 and check that upload to art2 works
        self.controller.delete(user1_req, 'sample_artifact', art1['id'])
        self.controller.upload_blob(
            user1_req, 'sample_artifact', art2['id'], 'blob',
            BytesIO(b'a' * 300), 'application/octet-stream', 300)

        # upload 700 bytes to img1 works
        self.controller.upload_blob(
            user1_req, 'images', img1['id'], 'image',
            BytesIO(b'a' * 700), 'application/octet-stream', 700)

        # upload to img2 fails because of global limit
        self.assertRaises(
            exception.RequestEntityTooLarge, self.controller.upload_blob,
            user1_req, 'images', img2['id'], 'image',
            BytesIO(b'a'), 'application/octet-stream', 1)

        # user2 can upload data to images
        img1 = self.controller.create(
            user2_req, 'images', {'name': 'img1'})
        self.controller.upload_blob(
            user2_req, 'images', img1['id'], 'image',
            BytesIO(b'a' * 1000), 'application/octet-stream', 1000)

        # disable global limit and try upload data from user1 again
        self.config(max_uploaded_data=-1)
        self.controller.upload_blob(
            user1_req, 'images', img2['id'], 'image',
            BytesIO(b'a' * 800), 'application/octet-stream', 800)

        # uploading more fails because of image type limit
        self.assertRaises(
            exception.RequestEntityTooLarge, self.controller.upload_blob,
            user1_req, 'images', img3['id'], 'image',
            BytesIO(b'a'), 'application/octet-stream', 1)

        # disable type limit and try upload data from user1 again
        self.config(max_uploaded_data=-1, group='artifact_type:images')
        self.controller.upload_blob(
            user1_req, 'images', img3['id'], 'image',
            BytesIO(b'a' * 1000), 'application/octet-stream', 1000)


class TestDynamicQuotas(base.BaseTestArtifactAPI):
    """Test dynamic quota limits."""

    def test_count_artifact_number(self):
        user1_req = self.get_fake_request(self.users['user1'])
        user2_req = self.get_fake_request(self.users['user2'])
        # initially there are no artifacts
        self.assertEqual(
            0, len(self.controller.list(user1_req, 'all')['artifacts']))
        self.assertEqual(
            0, len(self.controller.list(user2_req, 'all')['artifacts']))

        values = {
            user1_req.context.tenant: {
                "max_artifact_number:images": 3,
                "max_artifact_number:heat_templates": 15,
                "max_artifact_number:murano_packages": 10,
                "max_artifact_number": 10
            },
            user2_req.context.tenant: {
                "max_artifact_number": 10
            }
        }

        admin_req = self.get_fake_request(self.users["admin"])
        # define several quotas
        self.controller.set_quotas(admin_req, values)

        # create 3 images for user1
        for i in range(3):
            img = self.controller.create(
                user1_req, 'images', {'name': 'img%d' % i})

        # creation of another image fails because of artifact type limit
        self.assertRaises(exception.Forbidden, self.controller.create,
                          user1_req, 'images', {'name': 'img4'})

        # create 7 murano packages
        for i in range(7):
            self.controller.create(
                user1_req, 'murano_packages', {'name': 'mp%d' % i})

        # creation of another package fails because of global limit
        self.assertRaises(exception.Forbidden, self.controller.create,
                          user1_req, 'murano_packages', {'name': 'mp8'})

        # delete an image and create another murano package work
        self.controller.delete(user1_req, 'images', img['id'])
        self.controller.create(user1_req, 'murano_packages', {'name': 'mp8'})

        # user2 can create his own artifacts
        for i in range(10):
            self.controller.create(
                user2_req, 'heat_templates', {'name': 'ht%d' % i})

        # creation of another heat template fails because of global limit
        self.assertRaises(exception.Forbidden, self.controller.create,
                          user2_req, 'heat_templates', {'name': 'ht11'})

        # disable global limit for user1 and try to create 15 heat templates
        values = {
            user1_req.context.tenant: {
                "max_artifact_number:images": 3,
                "max_artifact_number:heat_templates": 15,
                "max_artifact_number:murano_packages": 10,
                "max_artifact_number": -1
            }
        }
        self.controller.set_quotas(admin_req, values)

        for i in range(15):
            self.controller.create(
                user1_req, 'heat_templates', {'name': 'ht%d' % i})

        # creation of another heat template fails because of type limit
        self.assertRaises(exception.Forbidden, self.controller.create,
                          user1_req, 'heat_templates', {'name': 'ht16'})

        # disable type limit for heat templates and create 1 heat templates
        values = {
            user1_req.context.tenant: {
                "max_artifact_number:images": 3,
                "max_artifact_number:heat_templates": -1,
                "max_artifact_number:murano_packages": 10,
                "max_artifact_number": -1
            }
        }
        self.controller.set_quotas(admin_req, values)

        # now user1 can create another heat template
        self.controller.create(
            user1_req, 'heat_templates', {'name': 'ht16'})

    def test_calculate_uploaded_data(self):
        user1_req = self.get_fake_request(self.users['user1'])
        user2_req = self.get_fake_request(self.users['user2'])
        # initially there are no artifacts
        self.assertEqual(
            0, len(self.controller.list(user1_req, 'all')['artifacts']))
        self.assertEqual(
            0, len(self.controller.list(user2_req, 'all')['artifacts']))

        values = {
            user1_req.context.tenant: {
                "max_uploaded_data:images": 1500,
                "max_uploaded_data:sample_artifact": 300,
                "max_uploaded_data:murano_packages": 1000,
                "max_uploaded_data": 1000
            },
            user2_req.context.tenant: {
                "max_uploaded_data": 1000
            }
        }

        admin_req = self.get_fake_request(self.users["admin"])
        # define several quotas
        self.controller.set_quotas(admin_req, values)
        # create 2 sample artifacts for user 1
        art1 = self.controller.create(
            user1_req, 'sample_artifact', {'name': 'art1'})
        art2 = self.controller.create(
            user1_req, 'sample_artifact', {'name': 'art2'})

        # create 3 images for user1
        img1 = self.controller.create(
            user1_req, 'images', {'name': 'img1'})
        img2 = self.controller.create(
            user1_req, 'images', {'name': 'img2'})
        img3 = self.controller.create(
            user1_req, 'images', {'name': 'img3'})

        # upload to art1 fails now because of type limit
        self.assertRaises(
            exception.RequestEntityTooLarge, self.controller.upload_blob,
            user1_req, 'sample_artifact', art1['id'], 'blob',
            BytesIO(b'a' * 301), 'application/octet-stream', 301)

        # upload to img1 fails now because of global limit
        self.assertRaises(
            exception.RequestEntityTooLarge, self.controller.upload_blob,
            user1_req, 'images', img1['id'], 'image',
            BytesIO(b'a' * 1001), 'application/octet-stream', 1001)

        # upload 300 bytes to 'blob' of art1
        self.controller.upload_blob(
            user1_req, 'sample_artifact', art1['id'], 'blob',
            BytesIO(b'a' * 300), 'application/octet-stream',
            content_length=300)

        # upload another blob to art1 fails because of type limit
        self.assertRaises(
            exception.RequestEntityTooLarge, self.controller.upload_blob,
            user1_req, 'sample_artifact', art1['id'],
            'dict_of_blobs/blob', BytesIO(b'a'),
            'application/octet-stream', 1)

        # upload to art2 fails now because of type limit
        self.assertRaises(
            exception.RequestEntityTooLarge, self.controller.upload_blob,
            user1_req, 'sample_artifact', art2['id'], 'blob',
            BytesIO(b'a'), 'application/octet-stream', 1)

        # delete art1 and check that upload to art2 works
        self.controller.delete(user1_req, 'sample_artifact', art1['id'])
        self.controller.upload_blob(
            user1_req, 'sample_artifact', art2['id'], 'blob',
            BytesIO(b'a' * 300), 'application/octet-stream', 300)

        # upload 700 bytes to img1 works
        self.controller.upload_blob(
            user1_req, 'images', img1['id'], 'image',
            BytesIO(b'a' * 700), 'application/octet-stream', 700)

        # upload to img2 fails because of global limit
        self.assertRaises(
            exception.RequestEntityTooLarge, self.controller.upload_blob,
            user1_req, 'images', img2['id'], 'image',
            BytesIO(b'a'), 'application/octet-stream', 1)

        # user2 can upload data to images
        img1 = self.controller.create(
            user2_req, 'images', {'name': 'img1'})
        self.controller.upload_blob(
            user2_req, 'images', img1['id'], 'image',
            BytesIO(b'a' * 1000), 'application/octet-stream', 1000)

        # disable global limit and try upload data from user1 again
        values = {
            user1_req.context.tenant: {
                "max_uploaded_data:images": 1500,
                "max_uploaded_data:sample_artifact": 300,
                "max_uploaded_data:murano_packages": 1000,
                "max_uploaded_data": -1
            }
        }
        self.controller.set_quotas(admin_req, values)

        self.controller.upload_blob(
            user1_req, 'images', img2['id'], 'image',
            BytesIO(b'a' * 800), 'application/octet-stream', 800)

        # uploading more fails because of image type limit
        self.assertRaises(
            exception.RequestEntityTooLarge, self.controller.upload_blob,
            user1_req, 'images', img3['id'], 'image',
            BytesIO(b'a'), 'application/octet-stream', 1)

        # disable type limit and try upload data from user1 again
        values = {
            user1_req.context.tenant: {
                "max_uploaded_data:images": -1,
                "max_uploaded_data:sample_artifact": 300,
                "max_uploaded_data:murano_packages": 1000,
                "max_uploaded_data": -1
            }
        }
        self.controller.set_quotas(admin_req, values)
        self.controller.upload_blob(
            user1_req, 'images', img3['id'], 'image',
            BytesIO(b'a' * 1000), 'application/octet-stream', 1000)

    def test_quota_upload_no_content_length(self):
        user1_req = self.get_fake_request(self.users['user1'])
        user2_req = self.get_fake_request(self.users['user2'])
        admin_req = self.get_fake_request(self.users['admin'])

        values = {
            user1_req.context.tenant: {
                "max_uploaded_data:sample_artifact": 20,
                "max_uploaded_data": 5
            },
            user2_req.context.tenant: {
                "max_uploaded_data:sample_artifact": 7,
                "max_uploaded_data": -1
            },
            admin_req.context.tenant: {
                "max_uploaded_data:sample_artifact": -1,
                "max_uploaded_data": -1
            }
        }

        # define several quotas
        self.controller.set_quotas(admin_req, values)

        # create a sample artifacts for user 1
        art1 = self.controller.create(
            user1_req, 'sample_artifact', {'name': 'art1'})

        # Max small_blob size is 10. User1 global quota is 5.
        # Since user doesn't specify how many bytes he wants to upload,
        # engine can't verify it before upload. Therefore it allocates
        # 5 available bytes for user and begins upload. If uploaded data
        # amount exceeds this limit RequestEntityTooLarge is raised and
        # upload fails.
        with mock.patch(
                'glare.common.store_api.save_blob_to_store',
                side_effect=store_api.save_blob_to_store) as mocked_save:
            data = BytesIO(b'a' * 10)
            self.assertRaises(
                exception.RequestEntityTooLarge,
                self.controller.upload_blob,
                user1_req, 'sample_artifact', art1['id'], 'small_blob',
                data, 'application/octet-stream',
                content_length=None)
            mocked_save.assert_called_once_with(
                mock.ANY, data, user1_req.context, 5, store_type='database')

        # check that blob wasn't uploaded
        self.assertIsNone(
            self.controller.show(
                user1_req, 'sample_artifact', art1['id'])['small_blob'])

        # try to upload with smaller amount that doesn't exceeds quota
        with mock.patch(
                'glare.common.store_api.save_blob_to_store',
                side_effect=store_api.save_blob_to_store) as mocked_save:
            data = BytesIO(b'a' * 4)
            self.controller.upload_blob(
                user1_req, 'sample_artifact', art1['id'], 'small_blob',
                data, 'application/octet-stream',
                content_length=None)
            mocked_save.assert_called_once_with(
                mock.ANY, data, user1_req.context, 5, store_type='database')

        # check that blob was uploaded
        blob = self.controller.show(
            user1_req, 'sample_artifact', art1['id'])['small_blob']
        self.assertEqual(4, blob['size'])
        self.assertEqual('active', blob['status'])

        # create a sample artifacts for user 2
        art2 = self.controller.create(
            user2_req, 'sample_artifact', {'name': 'art2'})

        # Max small_blob size is 10. User1 has no global quota, but his
        # type quota is 7.
        # Since user doesn't specify how many bytes he wants to upload,
        # engine can't verify it before upload. Therefore it allocates
        # 7 available bytes for user and begins upload. If uploaded data
        # amount exceeds this limit RequestEntityTooLarge is raised and
        # upload fails.
        with mock.patch(
                'glare.common.store_api.save_blob_to_store',
                side_effect=store_api.save_blob_to_store) as mocked_save:
            data = BytesIO(b'a' * 10)
            self.assertRaises(
                exception.RequestEntityTooLarge,
                self.controller.upload_blob,
                user2_req, 'sample_artifact', art2['id'], 'small_blob',
                data, 'application/octet-stream',
                content_length=None)
            mocked_save.assert_called_once_with(
                mock.ANY, data, user2_req.context, 7, store_type='database')

        # check that blob wasn't uploaded
        self.assertIsNone(
            self.controller.show(
                user2_req, 'sample_artifact', art2['id'])['small_blob'])

        # try to upload with smaller amount that doesn't exceeds quota
        with mock.patch(
                'glare.common.store_api.save_blob_to_store',
                side_effect=store_api.save_blob_to_store) as mocked_save:
            data = BytesIO(b'a' * 7)
            self.controller.upload_blob(
                user2_req, 'sample_artifact', art2['id'], 'small_blob',
                data, 'application/octet-stream',
                content_length=None)
            mocked_save.assert_called_once_with(
                mock.ANY, data, user2_req.context, 7, store_type='database')

        # check that blob was uploaded
        blob = self.controller.show(
            user2_req, 'sample_artifact', art2['id'])['small_blob']
        self.assertEqual(7, blob['size'])
        self.assertEqual('active', blob['status'])

        # create a sample artifacts for admin
        arta = self.controller.create(
            user2_req, 'sample_artifact', {'name': 'arta'})

        # Max small_blob size is 10. Admin has no quotas at all.
        # Since admin doesn't specify how many bytes he wants to upload,
        # engine can't verify it before upload. Therefore it allocates
        # 10 available bytes (max allowed small_blob size) for him and begins
        # upload. If uploaded data amount exceeds this limit
        # RequestEntityTooLarge is raised and upload fails.
        with mock.patch(
                'glare.common.store_api.save_blob_to_store',
                side_effect=store_api.save_blob_to_store) as mocked_save:
            data = BytesIO(b'a' * 11)
            self.assertRaises(
                exception.RequestEntityTooLarge,
                self.controller.upload_blob,
                admin_req, 'sample_artifact', arta['id'], 'small_blob',
                data, 'application/octet-stream',
                content_length=None)
            mocked_save.assert_called_once_with(
                mock.ANY, data, admin_req.context, 10, store_type='database')

        # check that blob wasn't uploaded
        self.assertIsNone(
            self.controller.show(
                admin_req, 'sample_artifact', arta['id'])['small_blob'])

        # try to upload with smaller amount that doesn't exceeds quota
        with mock.patch(
                'glare.common.store_api.save_blob_to_store',
                side_effect=store_api.save_blob_to_store) as mocked_save:
            data = BytesIO(b'a' * 10)
            self.controller.upload_blob(
                admin_req, 'sample_artifact', arta['id'], 'small_blob',
                data, 'application/octet-stream',
                content_length=None)
            mocked_save.assert_called_once_with(
                mock.ANY, data, admin_req.context, 10, store_type='database')

        # check that blob was uploaded
        blob = self.controller.show(
            admin_req, 'sample_artifact', arta['id'])['small_blob']
        self.assertEqual(10, blob['size'])
        self.assertEqual('active', blob['status'])



# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# Copyright 2014 SoftLayer Technologies, Inc.
# Copyright 2015 Mirantis, Inc
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
System-level utilities and helper functions.
"""
try:
    from eventlet import sleep
except ImportError:
    from time import sleep
from eventlet.green import socket

import hashlib
import os
import re

import glance_store
from OpenSSL import crypto
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import encodeutils
from oslo_utils import excutils
from oslo_utils import timeutils
from oslo_utils import uuidutils
from oslo_versionedobjects import fields
import requests
import six

from glare.common import exception
from glare.i18n import _
from glare.objects.meta import fields as glare_fields

CONF = cfg.CONF

LOG = logging.getLogger(__name__)

GLARE_TEST_SOCKET_FD_STR = 'GLARE_TEST_SOCKET_FD'


def cooperative_iter(iter):
    """Return an iterator which schedules after each
    iteration. This can prevent eventlet thread starvation.

    :param iter: an iterator to wrap
    """
    try:
        for chunk in iter:
            sleep(0)
            yield chunk
    except Exception as err:
        with excutils.save_and_reraise_exception():
            LOG.error("Error: cooperative_iter exception %s", err)


def cooperative_read(fd):
    """Wrap a file descriptor's read with a partial function which schedules
    after each read. This can prevent eventlet thread starvation.

    :param fd: a file descriptor to wrap
    """
    def readfn(*args):
        result = fd.read(*args)
        sleep(0)
        return result
    return readfn


MAX_COOP_READER_BUFFER_SIZE = 134217728  # 128M seems like a sane buffer limit


class CooperativeReader(object):
    """An eventlet thread friendly class for reading in blob data.

    When accessing data either through the iterator or the read method
    we perform a sleep to allow a co-operative yield. When there is more than
    one blob being uploaded/downloaded this prevents eventlet thread
    starvation, ie allows all threads to be scheduled periodically rather than
    having the same thread be continuously active.
    """
    def __init__(self, fd):
        """:param fd: Underlying blob file object
        """
        self.fd = fd
        self.iterator = None
        # NOTE(markwash): if the underlying supports read(), overwrite the
        # default iterator-based implementation with cooperative_read which
        # is more straightforward
        if hasattr(fd, 'read'):
            self.read = cooperative_read(fd)
        else:
            self.iterator = None
            self.buffer = b''
            self.position = 0

    def read(self, length=None):
        """Return the requested amount of bytes, fetching the next chunk of
        the underlying iterator when needed.

        This is replaced with cooperative_read in __init__ if the underlying
        fd already supports read().
        """
        if length is None:
            if len(self.buffer) - self.position > 0:
                # if no length specified but some data exists in buffer,
                # return that data and clear the buffer
                result = self.buffer[self.position:]
                self.buffer = b''
                self.position = 0
                return str(result)
            else:
                # otherwise read the next chunk from the underlying iterator
                # and return it as a whole. Reset the buffer, as subsequent
                # calls may specify the length
                try:
                    if self.iterator is None:
                        self.iterator = self.__iter__()
                    return next(self.iterator)
                except StopIteration:
                    return ''
                finally:
                    self.buffer = b''
                    self.position = 0
        else:
            result = bytearray()
            while len(result) < length:
                if self.position < len(self.buffer):
                    to_read = length - len(result)
                    chunk = self.buffer[self.position:self.position + to_read]
                    result.extend(chunk)

                    # This check is here to prevent potential OOM issues if
                    # this code is called with unreasonably high values of read
                    # size. Currently it is only called from the HTTP clients
                    # of Glare backend stores, which use httplib for data
                    # streaming, which has readsize hardcoded to 8K, so this
                    # check should never fire. Regardless it still worths to
                    # make the check, as the code may be reused somewhere else.
                    if len(result) >= MAX_COOP_READER_BUFFER_SIZE:
                        raise exception.RequestEntityTooLarge()
                    self.position += len(chunk)
                else:
                    try:
                        if self.iterator is None:
                            self.iterator = self.__iter__()
                        self.buffer = next(self.iterator)
                        self.position = 0
                    except StopIteration:
                        self.buffer = b''
                        self.position = 0
                        return bytes(result)
            return bytes(result)

    def __iter__(self):
        return cooperative_iter(self.fd.__iter__())


class LimitingReader(object):
    """Reader designed to fail when reading blob data past the configured
    allowable amount.
    """
    def __init__(self, data, limit):
        """
        :param data: Underlying blob data object
        :param limit: maximum number of bytes the reader should allow
        """
        self.data = data
        self.limit = limit
        self.bytes_read = 0
        self.md5 = hashlib.md5()
        self.sha1 = hashlib.sha1()
        self.sha256 = hashlib.sha256()

    def __iter__(self):
        for chunk in self.data:
            self.bytes_read += len(chunk)
            if self.bytes_read > self.limit:
                raise exception.RequestEntityTooLarge()
            else:
                yield chunk

    def read(self, length=None):
        res = self.data.read() if length is None else self.data.read(length)
        len_result = len(res)
        self.bytes_read += len_result
        if len_result:
            self.md5.update(res)
            self.sha1.update(res)
            self.sha256.update(res)
        if self.bytes_read > self.limit:
            message = _("The server is refusing to process a request because"
                        " the request entity is larger than the server is"
                        " willing or able to process - %s bytes.") % self.limit
            raise exception.RequestEntityTooLarge(message=message)
        return res


def validate_key_cert(key_file, cert_file):
    try:
        error_key_name = "private key"
        error_filename = key_file
        with open(key_file, 'r') as keyfile:
            key_str = keyfile.read()
        key = crypto.load_privatekey(crypto.FILETYPE_PEM, key_str)

        error_key_name = "certificate"
        error_filename = cert_file
        with open(cert_file, 'r') as certfile:
            cert_str = certfile.read()
        cert = crypto.load_certificate(crypto.FILETYPE_PEM, cert_str)
    except IOError as ioe:
        raise RuntimeError(_("There is a problem with your %(error_key_name)s "
                             "%(error_filename)s.  Please verify it."
                             "  Error: %(ioe)s") %
                           {'error_key_name': error_key_name,
                            'error_filename': error_filename,
                            'ioe': ioe})
    except crypto.Error as ce:
        raise RuntimeError(_("There is a problem with your %(error_key_name)s "
                             "%(error_filename)s.  Please verify it. OpenSSL"
                             " error: %(ce)s") %
                           {'error_key_name': error_key_name,
                            'error_filename': error_filename,
                            'ce': ce})

    try:
        data = uuidutils.generate_uuid()
        # On Python 3, explicitly encode to UTF-8 to call crypto.sign() which
        # requires bytes. Otherwise, it raises a deprecation warning (and
        # will raise an error later).
        data = encodeutils.to_utf8(data)
        digest = CONF.digest_algorithm
        if digest == 'sha1':
            LOG.warning(
                'The FIPS (FEDERAL INFORMATION PROCESSING STANDARDS)'
                ' state that the SHA-1 is not suitable for'
                ' general-purpose digital signature applications (as'
                ' specified in FIPS 186-3) that require 112 bits of'
                ' security. The default value is sha1 in Kilo for a'
                ' smooth upgrade process, and it will be updated'
                ' with sha256 in next release(L).')
        out = crypto.sign(key, data, digest)
        crypto.verify(cert, out, data, digest)
    except crypto.Error as ce:
        raise RuntimeError(_("There is a problem with your key pair.  "
                             "Please verify that cert %(cert_file)s and "
                             "key %(key_file)s belong together.  OpenSSL "
                             "error %(ce)s") % {'cert_file': cert_file,
                                                'key_file': key_file,
                                                'ce': ce})


def get_test_suite_socket():
    global GLARE_TEST_SOCKET_FD_STR
    if GLARE_TEST_SOCKET_FD_STR in os.environ:
        fd = int(os.environ[GLARE_TEST_SOCKET_FD_STR])
        sock = socket.fromfd(fd, socket.AF_INET, socket.SOCK_STREAM)
        if six.PY2:
            sock = socket.SocketType(_sock=sock)
        sock.listen(CONF.backlog)
        del os.environ[GLARE_TEST_SOCKET_FD_STR]
        os.close(fd)
        return sock
    return None


try:
    REGEX_4BYTE_UNICODE = re.compile(u'[\U00010000-\U0010ffff]')
except re.error:
    # UCS-2 build case
    REGEX_4BYTE_UNICODE = re.compile(u'[\uD800-\uDBFF][\uDC00-\uDFFF]')


def no_4byte_params(f):
    """Checks that no 4 byte unicode characters are allowed
    in dicts' keys/values and string's parameters.
    """
    def wrapper(*args, **kwargs):

        def _is_match(some_str):
            return (isinstance(some_str, six.text_type) and
                    REGEX_4BYTE_UNICODE.findall(some_str) != [])

        def _check_dict(data_dict):
            # a dict of dicts has to be checked recursively
            for key, value in data_dict.items():
                if isinstance(value, dict):
                    _check_dict(value)
                else:
                    if _is_match(key):
                        msg = _("Property names can't contain 4 byte unicode.")
                        raise exception.BadRequest(msg)
                    if _is_match(value):
                        msg = (_("%s can't contain 4 byte unicode characters.")
                               % key.title())
                        raise exception.BadRequest(msg)

        for data_dict in [arg for arg in args if isinstance(arg, dict)]:
            _check_dict(data_dict)
        # now check args for str values
        for arg in args:
            if _is_match(arg):
                msg = _("Param values can't contain 4 byte unicode.")
                raise exception.BadRequest(msg)
        # check kwargs as well, as params are passed as kwargs via
        # registry calls
        _check_dict(kwargs)
        return f(*args, **kwargs)
    return wrapper


def stash_conf_values():
    """Make a copy of some of the current global CONF's settings.
    Allows determining if any of these values have changed
    when the config is reloaded.
    """
    conf = {
        'bind_host': CONF.bind_host,
        'bind_port': CONF.bind_port,
        'tcp_keepidle': CONF.cert_file,
        'backlog': CONF.backlog,
        'key_file': CONF.key_file,
        'cert_file': CONF.cert_file,
        'enabled_artifact_types': CONF.enabled_artifact_types,
        'custom_artifact_types_modules': CONF.custom_artifact_types_modules
    }

    return conf


def split_filter_op(expression):
    """Split operator from threshold in an expression.
    Designed for use on a comparative-filtering query field.
    When no operator is found, default to an equality comparison.

    :param expression: the expression to parse
    :return: a tuple (operator, threshold) parsed from expression
    """
    left, sep, right = expression.partition(':')
    if sep:
        # If the expression is a date of the format ISO 8601 like
        # CCYY-MM-DDThh:mm:ss+hh:mm and has no operator, it should
        # not be partitioned, and a default operator of eq should be
        # assumed.
        try:
            timeutils.parse_isotime(expression)
            op = 'eq'
            threshold = expression
        except ValueError:
            op = left
            threshold = right
    else:
        op = 'eq'  # default operator
        threshold = left

    # NOTE stevelle decoding escaped values may be needed later
    return op, threshold


def validate_quotes(value):
    """Validate filter values

    Validation opening/closing quotes in the expression.
    """
    open_quotes = True
    for i in range(len(value)):
        if value[i] == '"':
            if i and value[i - 1] == '\\':
                continue
            if open_quotes:
                if i and value[i - 1] != ',':
                    msg = _("Invalid filter value %s. There is no comma "
                            "before opening quotation mark.") % value
                    raise exception.InvalidParameterValue(message=msg)
            else:
                if i + 1 != len(value) and value[i + 1] != ",":
                    msg = _("Invalid filter value %s. There is no comma "
                            "after closing quotation mark.") % value
                    raise exception.InvalidParameterValue(message=msg)
            open_quotes = not open_quotes
    if not open_quotes:
        msg = _("Invalid filter value %s. The quote is not closed.") % value
        raise exception.InvalidParameterValue(message=msg)


def split_filter_value_for_quotes(value):
    """Split filter values

    Split values by commas and quotes for 'in' operator, according api-wg.
    """
    validate_quotes(value)
    tmp = re.compile(r'''
        "(                 # if found a double-quote
           [^\"\\]*        # take characters either non-quotes or backslashes
           (?:\\.          # take backslashes and character after it
            [^\"\\]*)*     # take characters either non-quotes or backslashes
         )                 # before double-quote
        ",?                # a double-quote with comma maybe
        | ([^,]+),?        # if not found double-quote take any non-comma
                           # characters with comma maybe
        | ,                # if we have only comma take empty string
        ''', re.VERBOSE)
    return [val[0] or val[1] for val in re.findall(tmp, value)]


class error_handler(object):
    def __init__(self, error_map, default_exception=None):
        """Init method of the class.

        :param error_map: dict of exception that can be raised
        in func and exceptions that must be raised for these exceptions.
        For example, if sqlalchemy NotFound might be raised and we need
        re-raise it as glare NotFound exception then error_map must
        contain {"catch": SQLAlchemyNotFound,
                 "raise": exceptions.NotFound}
        :param default_exception: default exception that must be raised if
        exception that cannot be found in error map was raised
        :return: func
        """
        self.error_map = error_map
        self.default_exception = default_exception

    def __call__(self, f):
        """Decorator that catches exception that came from func or method.

        :param f: target func
        """

        def new_function(*args, **kwargs):
            try:
                return f(*args, **kwargs)
            except Exception as e:
                for map_record in self.error_map:
                    if isinstance(e, map_record['catch']):
                        raise map_record['raise'](str(e))
                else:
                    if self.default_exception:
                        raise self.default_exception(str(e))
                    else:
                        raise
        return new_function


def get_schema_type(attr):
    if isinstance(attr, fields.IntegerField) or attr is fields.Integer:
        return 'integer'
    elif isinstance(attr, fields.FloatField) or attr is fields.Float:
        return 'number'
    elif isinstance(attr, fields.FlexibleBooleanField) \
            or attr is fields.FlexibleBoolean:
        return 'boolean'
    elif isinstance(attr, glare_fields.List):
        return 'array'
    elif isinstance(attr, (glare_fields.Dict, glare_fields.BlobField)):
        return 'object'
    return 'string'


def get_glare_type(attr):
    if isinstance(attr, fields.IntegerField):
        return 'Integer'
    elif isinstance(attr, fields.FloatField):
        return 'Float'
    elif isinstance(attr, fields.FlexibleBooleanField):
        return 'Boolean'
    elif isinstance(attr, fields.DateTimeField):
        return 'DateTime'
    elif isinstance(attr, glare_fields.BlobField):
        return 'Blob'
    elif isinstance(attr, glare_fields.Link):
        return 'Link'
    elif isinstance(attr, glare_fields.List):
        return _get_element_type(attr.element_type) + 'List'
    elif isinstance(attr, glare_fields.Dict):
        return _get_element_type(attr.element_type) + 'Dict'
    return 'String'


def _get_element_type(element_type):
    if element_type is fields.FlexibleBooleanField:
        return 'Boolean'
    elif element_type is fields.Integer:
        return 'Integer'
    elif element_type is fields.Float:
        return 'Float'
    elif element_type is glare_fields.BlobFieldType:
        return 'Blob'
    elif element_type is glare_fields.LinkFieldType:
        return 'Link'
    return 'String'


class BlobIterator(object):
    """Reads data from a blob, one chunk at a time.
    """

    def __init__(self, data, chunk_size=65536):
        self.chunk_size = chunk_size
        self.data = data

    def __iter__(self):
        bytes_left = len(self.data)
        i = 0
        while bytes_left > 0:
            data = self.data[i * self.chunk_size:(i + 1) * self.chunk_size]
            bytes_left -= len(data)
            yield data
        raise StopIteration()


def validate_status_transition(af, from_status, to_status):
    if from_status == 'deleted':
        msg = _("Cannot change status if artifact is deleted.")
        raise exception.Forbidden(msg)
    if to_status == 'active':
        if from_status == 'drafted':
            for name, type_obj in af.fields.items():
                if type_obj.required_on_activate and getattr(af, name) is None:
                    msg = _("'%s' field value must be set before "
                            "activation.") % name
                    raise exception.Forbidden(msg)
    elif to_status == 'drafted':
        if from_status != 'drafted':
            msg = _("Cannot change status to 'drafted'") % from_status
            raise exception.Forbidden(msg)
    elif to_status == 'deactivated':
        if from_status not in ('active', 'deactivated'):
            msg = _("Cannot deactivate artifact if it's not active.")
            raise exception.Forbidden(msg)
    elif to_status == 'deleted':
        msg = _("Cannot delete artifact with PATCH requests. Use special "
                "API to do this.")
        raise exception.Forbidden(msg)
    else:
        msg = _("Unknown artifact status: %s.") % to_status
        raise exception.BadRequest(msg)


def validate_visibility_transition(af, from_visibility, to_visibility):
    if to_visibility == 'private':
        if from_visibility != 'private':
            msg = _("Cannot make artifact private again.")
            raise exception.Forbidden()
    elif to_visibility == 'public':
        if af.status != 'active':
            msg = _("Cannot change visibility to 'public' if artifact"
                    " is not active.")
            raise exception.Forbidden(msg)
    else:
        msg = _("Unknown artifact visibility: %s.") % to_visibility
        raise exception.BadRequest(msg)


def validate_change_allowed(af, field_name):
    """Validate if fields can be set for the artifact."""
    if field_name not in af.fields:
        msg = _("Cannot add new field '%s' to artifact.") % field_name
        raise exception.BadRequest(msg)
    if af.status not in ('active', 'drafted'):
        msg = _("Forbidden to change fields "
                "if artifact is not active or drafted.")
        raise exception.Forbidden(message=msg)
    if af.fields[field_name].system is True:
        msg = _("Forbidden to specify system field %s. It is not "
                "available for modifying by users.") % field_name
        raise exception.Forbidden(msg)
    if af.status == 'active' and not af.fields[field_name].mutable:
        msg = (_("Forbidden to change field '%s' after activation.")
               % field_name)
        raise exception.Forbidden(message=msg)


def initialize_glance_store():
    """Initialize glance store."""
    glance_store.register_opts(CONF)
    set_glance_store_config_defaults()
    glance_store.create_stores(CONF)
    glance_store.verify_default_store()


def set_glance_store_config_defaults():
    # By default glance and glare share common place to store data.
    # To prevent possible collisions we have to set other glance_store default
    # values for various backends.
    cfg.set_defaults(glance_store.backend._list_opts()[1][1],
                     # default '/var/lib/glance/images
                     filesystem_store_datadir='/var/lib/glare/artifacts',
                     # default 'images'
                     rbd_store_pool='artifacts',
                     # default '/openstack_glance
                     vmware_store_image_dir='/openstack_glare',
                     # default 'glance'
                     swift_store_container='glare')


def get_system_ca_file():
    """Return path to system default CA file."""
    # Standard CA file locations for Debian/Ubuntu, RedHat/Fedora,
    # Suse, FreeBSD/OpenBSD, MacOSX, and the bundled ca
    ca_path = ['/etc/ssl/certs/ca-certificates.crt',
               '/etc/pki/tls/certs/ca-bundle.crt',
               '/etc/ssl/ca-bundle.pem',
               '/etc/ssl/cert.pem',
               '/System/Library/OpenSSL/certs/cacert.pem',
               requests.certs.where()]
    for ca in ca_path:
        LOG.debug("Looking for ca file %s", ca)
        if os.path.exists(ca):
            LOG.debug("Using ca file %s", ca)
            return ca
    LOG.warning("System ca file could not be found.")



#!/usr/bin/env python
#
# Copyright (c) 2016 Mirantis, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


"""
Glare (Glare Artifact Repository) API service.
"""

import os
import sys

import eventlet
from oslo_utils import encodeutils

eventlet.patcher.monkey_patch(all=False, socket=True, time=True,
                              select=True, thread=True, os=True, MySQLdb=True)

# If ../glare/__init__.py exists, add ../ to Python search path, so that
# it will override what happens to be installed in /usr/(local/)lib/python...
possible_topdir = os.path.normpath(os.path.join(os.path.abspath(sys.argv[0]),
                                   os.pardir,
                                   os.pardir))
if os.path.exists(os.path.join(possible_topdir, 'glare', '__init__.py')):
    sys.path.insert(0, possible_topdir)

import glance_store
from oslo_config import cfg
from oslo_log import log as logging
from osprofiler import initializer

from glare.common import config
from glare.common import exception
from glare.common import wsgi
from glare import notification


CONF = cfg.CONF
CONF.import_group("profiler", "glare.common.wsgi")
logging.register_options(CONF)

KNOWN_EXCEPTIONS = (RuntimeError,
                    exception.WorkerCreationFailure,
                    glance_store.exceptions.BadStoreConfiguration)


def fail(e):
    global KNOWN_EXCEPTIONS
    return_code = KNOWN_EXCEPTIONS.index(type(e)) + 1
    sys.stderr.write("ERROR: %s\n" % encodeutils.exception_to_unicode(e))
    sys.exit(return_code)


def main():
    try:
        config.parse_args()
        wsgi.set_eventlet_hub()
        logging.setup(CONF, 'glare')
        notification.set_defaults()

        if CONF.profiler.enabled:
            initializer.init_from_conf(
                conf=CONF,
                context={},
                project="glare",
                service="api",
                host=CONF.bind_host
            )

        server = wsgi.Server(initialize_glance_store=True)
        server.start(config.load_paste_app('glare-api'), default_port=9494)
        server.wait()
    except KNOWN_EXCEPTIONS as e:
        fail(e)


if __name__ == '__main__':
    main()



# Copyright 2017 - Nokia Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import os
from time import time

from glare.tests.unit import base


class TestArtifactHooks(base.BaseTestArtifactAPI):

    def setUp(self):
        super(TestArtifactHooks, self).setUp()
        values = {'name': 'ttt', 'version': '1.0'}
        self.unpacking_artifact = self.controller.create(
            self.req, 'unpacking_artifact', values)

    def test_unpacking(self):
        var_dir = os.path.abspath(os.path.join(os.path.dirname(__file__),
                                               '../', 'var'))
        data_path = os.path.join(var_dir, 'hooks.zip')
        with open(data_path, "rb") as data:
            self.controller.upload_blob(
                self.req, 'unpacking_artifact', self.unpacking_artifact['id'],
                'zip', data, 'application/octet-stream')
        artifact = self.controller.show(self.req, 'unpacking_artifact',
                                        self.unpacking_artifact['id'])
        self.assertEqual(818, artifact['zip']['size'])
        self.assertEqual('active', artifact['zip']['status'])

        self.assertEqual(11, artifact['content']['aaa.txt']['size'])
        self.assertEqual(11, artifact['content']['folder1/bbb.txt']['size'])
        self.assertEqual(
            11, artifact['content']['folder1/folder2/ccc.txt']['size'])

    def test_unpacking_database(self):
        self.config(default_store='database',
                    group='artifact_type:unpacking_artifact')
        self.test_unpacking()

    def test_unpacking_big_archive(self):
        var_dir = os.path.abspath(os.path.join(os.path.dirname(__file__),
                                               '../', 'var'))
        data_path = os.path.join(var_dir, 'hooks_100.zip')

        # play rally - test that this test should pass faster than 3 seconds
        start = time()
        with open(data_path, "rb") as data:
            self.controller.upload_blob(
                self.req, 'unpacking_artifact', self.unpacking_artifact['id'],
                'zip', data, 'application/octet-stream')
        end = time()
        self.assertIs(True, (end - start) < 3, (end - start))

        artifact = self.controller.show(self.req, 'unpacking_artifact',
                                        self.unpacking_artifact['id'])
        self.assertEqual(15702, artifact['zip']['size'])
        self.assertEqual('active', artifact['zip']['status'])

        self.assertEqual(100, len(artifact['content']))

        for blob in artifact['content'].values():
            self.assertEqual('active', blob['status'])
            self.assertEqual(15, blob['size'])

    def test_unpacking_database_big_archive(self):
        self.config(default_store='database',
                    group='artifact_type:unpacking_artifact')
        self.test_unpacking_big_archive()



# Copyright 2016 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from copy import deepcopy

from eventlet import tpool
import jsonpatch
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import excutils
from oslo_utils import timeutils
from oslo_utils import uuidutils
import six.moves.urllib.parse as urlparse

from glare.common import exception
from glare.common import policy
from glare.common import store_api
from glare.common import utils
from glare.db import artifact_api
from glare.i18n import _
from glare import locking
from glare.notification import Notifier
from glare.objects.meta import registry
from glare import quota

CONF = cfg.CONF
LOG = logging.getLogger(__name__)


class Engine(object):
    """Engine is responsible for executing different helper operations when
    processing incoming requests from Glare API.

    Engine receives incoming data and does the following:
     - check basic policy permissions;
     - requests artifact definition from artifact type registry;
     - check access permission(ro, rw);
     - lock artifact for update if needed;
     - pass data to base artifact type to execute all business logic operations
       with database;
     - check quotas during upload;
     - call operations pre- and post- hooks;
     - notify other users about finished operation.

    Engine should not include any business logic and validation related
    to artifacts types. Engine should not know any internal details of artifact
    type, because this part of the work is done by Base artifact type.
    """
    def __init__(self):
        # register all artifact types
        registry.ArtifactRegistry.register_all_artifacts()

        # generate all schemas and quotas
        self.schemas = {}
        self.config_quotas = {
            'max_artifact_number': CONF.max_artifact_number,
            'max_uploaded_data': CONF.max_uploaded_data
        }
        for name, type_list in registry.ArtifactRegistry.obj_classes().items():
            type_name = type_list[0].get_type_name()
            self.schemas[type_name] = registry.ArtifactRegistry.\
                get_artifact_type(type_name).gen_schemas()
            type_conf_section = getattr(CONF, 'artifact_type:' + type_name)
            if type_conf_section.max_artifact_number is not None:
                self.config_quotas['max_artifact_number:' + type_name] = \
                    type_conf_section.max_artifact_number
            if type_conf_section.max_uploaded_data is not None:
                self.config_quotas['max_uploaded_data:' + type_name] = \
                    type_conf_section.max_uploaded_data

    lock_engine = locking.LockEngine(artifact_api.ArtifactLockApi())

    def _create_scoped_lock(self, context, type_name, name, version,
                            owner, visibility='private'):
        """Create scoped lock for artifact."""
        # validate that artifact doesn't exist for the scope
        filters = [('name', 'eq:' + name), ('version', 'eq:' + version)]
        if visibility == 'public':
            filters.extend([('visibility', 'public')])
        elif visibility == 'private':
            filters.extend([('owner', 'eq:' + owner),
                            ('visibility', 'private')])

        scope_id = owner
        if visibility == 'public':
            scope_id = "%s:%s:%s" % (type_name, name, version)

        lock = self.lock_engine.acquire(context, scope_id)

        try:
            if len(self.list(context, type_name, filters)) > 0:
                msg = _("Artifact with this name and version is already "
                        "exists for this scope.")
                raise exception.Conflict(msg)
        except Exception:
            with excutils.save_and_reraise_exception(logger=LOG):
                self.lock_engine.release(lock)

        return lock

    @staticmethod
    def _show_artifact(ctx, type_name, artifact_id, read_only=False):
        """Return artifact requested by user.

        Check access permissions and policies.

        :param ctx: user context
        :param type_name: artifact type name
        :param artifact_id: id of the artifact to be updated
        :param read_only: flag, if set to True only read access is checked,
         if False then engine checks if artifact can be modified by the user
        """
        artifact_type = registry.ArtifactRegistry.get_artifact_type(type_name)
        # only artifact is available for class users
        af = artifact_type.show(ctx, artifact_id)
        if not read_only:
            if not ctx.is_admin and ctx.tenant != af.owner or ctx.read_only:
                raise exception.Forbidden()
            LOG.debug("Artifact %s acquired for read-write access",
                      artifact_id)
        else:
            LOG.debug("Artifact %s acquired for read-only access", artifact_id)

        return af

    def show_type_schemas(self, context, type_name=None):
        policy.authorize("artifact:type_list", {}, context)
        if type_name is None:
            return self.schemas
        if type_name not in self.schemas:
            msg = _("Artifact type %s does not exist") % type_name
            raise exception.NotFound(message=msg)
        return self.schemas[type_name]

    def _apply_patch(self, context, af, patch):
        # This function is a collection of hacks and workarounds to make
        # json patch apply changes to artifact object.
        action_names = ['update']
        af_dict = af.to_dict()
        policy.authorize('artifact:update', af_dict, context)
        af.pre_update_hook(context, af)
        try:
            for operation in patch._ops:
                # apply the change to make sure that it's correct
                af_dict = operation.apply(af_dict)

                # format of location is "/key/value" or just "/key"
                # first case symbolizes that we have dict or list insertion,
                # second, that we work with a field itself.
                items = operation.location.split('/', 2)
                field_name = items[1]
                if af.is_blob(field_name) or af.is_blob_dict(field_name):
                    msg = _("Cannot add blob with this request. "
                            "Use special Blob API for that.")
                    raise exception.BadRequest(msg)
                if len(items) == 2 and operation.operation['op'] == 'remove':
                    msg = _("Cannot remove field '%s' from "
                            "artifact.") % field_name
                    raise exception.BadRequest(msg)

                # work with hooks and define action names
                if field_name == 'visibility':
                    utils.validate_visibility_transition(
                        af,
                        from_visibility=af.visibility,
                        to_visibility=af_dict['visibility']
                    )
                    if af_dict['visibility'] == 'public':
                        policy.authorize(
                            'artifact:publish', af_dict, context)
                        af.pre_publish_hook(context, af)
                        action_names.append('publish')
                elif field_name == 'status':
                    utils.validate_status_transition(
                        af, from_status=af.status, to_status=af_dict['status'])
                    if af_dict['status'] == 'deactivated':
                        policy.authorize(
                            'artifact:deactivate', af_dict, context)
                        af.pre_deactivate_hook(context, af)
                        action_names.append('deactivate')
                    elif af_dict['status'] == 'active':
                        if af.status == 'deactivated':
                            policy.authorize(
                                'artifact:reactivate', af_dict, context)
                            af.pre_reactivate_hook(context, af)
                            action_names.append('reactivate')
                        else:
                            policy.authorize(
                                'artifact:activate', af_dict, context)
                            af.pre_activate_hook(context, af)
                            action_names.append('activate')
                else:
                    utils.validate_change_allowed(af, field_name)

                old_val = getattr(af, field_name)
                setattr(af, field_name, af_dict[field_name])
                new_val = getattr(af, field_name)
                if new_val == old_val:
                    # No need to save value to db if it's not changed
                    af.obj_reset_changes([field_name])

        except (jsonpatch.JsonPatchException,
                jsonpatch.JsonPointerException, TypeError) as e:
            raise exception.BadRequest(message=str(e))

        return action_names

    def create(self, context, type_name, values):
        """Create artifact record in Glare.

        :param context: user context
        :param type_name: artifact type name
        :param values: dict with artifact fields
        :return: dict representation of created artifact
        """
        action_name = "artifact:create"
        policy.authorize(action_name, values, context)
        artifact_type = registry.ArtifactRegistry.get_artifact_type(type_name)
        version = values.get('version', artifact_type.DEFAULT_ARTIFACT_VERSION)
        init_values = {
            'id': uuidutils.generate_uuid(),
            'name': values.pop('name'),
            'version': version,
            'owner': context.tenant,
            'created_at': timeutils.utcnow(),
            'updated_at': timeutils.utcnow()
        }
        af = artifact_type.init_artifact(context, init_values)
        # acquire scoped lock and execute artifact create
        with self._create_scoped_lock(context, type_name, af.name,
                                      af.version, context.tenant):
            quota.verify_artifact_count(context, type_name)
            for field_name, value in values.items():
                if af.is_blob(field_name) or af.is_blob_dict(field_name):
                    msg = _("Cannot add blob with this request. "
                            "Use special Blob API for that.")
                    raise exception.BadRequest(msg)
                utils.validate_change_allowed(af, field_name)
                setattr(af, field_name, value)
            artifact_type.pre_create_hook(context, af)
            af = af.create(context)
            artifact_type.post_create_hook(context, af)
            # notify about new artifact
            Notifier.notify(context, action_name, af)
            # return artifact to the user
            return af.to_dict()

    def save(self, context, type_name, artifact_id, patch):
        """Update artifact with json patch.

        Apply patch to artifact and validate artifact before updating it
        in database. If there is request for visibility or status change
        then call specific method for that.

        :param context: user context
        :param type_name: name of artifact type
        :param artifact_id: id of the artifact to be updated
        :param patch: json patch object
        :return: dict representation of updated artifact
        """
        lock_key = "%s:%s" % (type_name, artifact_id)
        with self.lock_engine.acquire(context, lock_key):
            af = self._show_artifact(context, type_name, artifact_id)
            af.obj_reset_changes()
            action_names = self._apply_patch(context, af, patch)
            updates = af.obj_changes_to_primitive()

            LOG.debug("Update diff successfully calculated for artifact "
                      "%(af)s %(diff)s", {'af': artifact_id, 'diff': updates})
            if not updates:
                return af.to_dict()

            if any(i in updates for i in ('name', 'version', 'visibility')):
                # to change an artifact scope it's required to set a lock first
                with self._create_scoped_lock(
                        context, type_name, updates.get('name', af.name),
                        updates.get('version', af.version), af.owner,
                        updates.get('visibility', af.visibility)):
                    af = af.save(context)
            else:
                af = af.save(context)

            # call post hooks for all operations when data is written in db and
            # send broadcast notifications
            for action_name in action_names:
                getattr(af, 'post_%s_hook' % action_name)(context, af)
                Notifier.notify(context, 'artifact:' + action_name, af)

            return af.to_dict()

    def show(self, context, type_name, artifact_id):
        """Show detailed artifact info.

        :param context: user context
        :param type_name: Artifact type name
        :param artifact_id: id of artifact to show
        :return: definition of requested artifact
        """
        policy.authorize("artifact:get", {}, context)
        af = self._show_artifact(context, type_name, artifact_id,
                                 read_only=True)
        return af.to_dict()

    @staticmethod
    def list(context, type_name, filters, marker=None, limit=None,
             sort=None, latest=False):
        """Return list of artifacts requested by user.

        :param context: user context
        :param type_name: Artifact type name
        :param filters: filters that need to be applied to artifact
        :param marker: the artifact that considered as begin of the list
         so all artifacts before marker (including marker itself) will not be
         added to artifact list
        :param limit: maximum number of items in list
        :param sort: sorting options
        :param latest: flag that indicates, that only artifacts with highest
         versions should be returned in output
        :return: list of artifact definitions
        """
        policy.authorize("artifact:list", {}, context)
        artifact_type = registry.ArtifactRegistry.get_artifact_type(type_name)
        # return list to the user
        af_list = [af.to_dict()
                   for af in artifact_type.list(context, filters, marker,
                                                limit, sort, latest)]
        return af_list

    @staticmethod
    def _delete_blobs(context, af, blobs):
        for name, blob in blobs.items():
            if af.is_blob(name):
                if not blob['external']:
                    try:
                        store_api.delete_blob(blob['url'], context=context)
                    except exception.NotFound:
                        # data has already been removed
                        pass
                af.db_api.update_blob(context, af.id, {name: None})
            elif af.is_blob_dict(name):
                upd_blob = deepcopy(blob)
                for key, val in blob.items():
                    if not val['external']:
                        try:
                            store_api.delete_blob(val['url'], context=context)
                        except exception.NotFound:
                            pass
                    del upd_blob[key]
                    af.db_api.update_blob(context, af.id, {name: upd_blob})

    def delete(self, context, type_name, artifact_id):
        """Delete artifact from Glare.

        :param context: User context
        :param type_name: Artifact type name
        :param artifact_id: id of artifact to delete
        """
        af = self._show_artifact(context, type_name, artifact_id)
        action_name = 'artifact:delete'
        policy.authorize(action_name, af.to_dict(), context)
        af.pre_delete_hook(context, af)
        blobs = af.delete(context, af)

        delayed_delete = getattr(
            CONF, 'artifact_type:' + type_name).delayed_delete
        # use global parameter if delayed delete isn't set per artifact type
        if delayed_delete is None:
            delayed_delete = CONF.delayed_delete

        if not delayed_delete:
            if blobs:
                # delete blobs one by one
                self._delete_blobs(context, af, blobs)
                LOG.info("Blobs successfully deleted for artifact %s", af.id)
            # delete artifact itself
            af.db_api.delete(context, af.id)
        af.post_delete_hook(context, af)
        Notifier.notify(context, action_name, af)

    @staticmethod
    def _get_blob_info(af, field_name, blob_key=None):
        """Return requested blob info."""
        if blob_key:
            if not af.is_blob_dict(field_name):
                msg = _("%s is not a blob dict") % field_name
                raise exception.BadRequest(msg)
            return getattr(af, field_name).get(blob_key)
        else:
            if not af.is_blob(field_name):
                msg = _("%s is not a blob") % field_name
                raise exception.BadRequest(msg)
            return getattr(af, field_name, None)

    @staticmethod
    def _save_blob_info(context, af, field_name, blob_key, value):
        """Save blob instance in database."""
        if blob_key is not None:
            # Insert blob value in the folder
            folder = getattr(af, field_name)
            if value is not None:
                folder[blob_key] = value
            else:
                del folder[blob_key]
            value = folder
        return af.update_blob(context, af.id, field_name, value)

    @staticmethod
    def _generate_blob_name(field_name, blob_key=None):
        return "%s[%s]" % (field_name, blob_key) if blob_key else field_name

    def add_blob_location(self, context, type_name, artifact_id, field_name,
                          location, blob_meta, blob_key=None):
        """Add external/internal location to blob.

        :param context: user context
        :param type_name: name of artifact type
        :param artifact_id: id of the artifact to be updated
        :param field_name: name of blob or blob dict field
        :param location: blob url
        :param blob_meta: dictionary containing blob metadata like md5 checksum
        :param blob_key: if field_name is blob dict it specifies key
         in this dict
        :return: dict representation of updated artifact
        """
        blob_name = self._generate_blob_name(field_name, blob_key)

        location_type = blob_meta.pop('location_type', 'external')

        if location_type == 'external':
            action_name = 'artifact:set_location'
        elif location_type == 'internal':
            scheme = urlparse.urlparse(location).scheme
            if scheme in store_api.RESTRICTED_URI_SCHEMES:
                msg = _("Forbidden to set internal locations with "
                        "scheme '%s'") % scheme
                raise exception.Forbidden(msg)
            if scheme not in store_api.get_known_schemes():
                msg = _("Unknown scheme '%s'") % scheme
                raise exception.BadRequest(msg)
            action_name = 'artifact:set_internal_location'
        else:
            msg = _("Invalid location type: %s") % location_type
            raise exception.BadRequest(msg)

        blob = {'url': location, 'size': None, 'md5': blob_meta.get("md5"),
                'sha1': blob_meta.get("sha1"), 'id': uuidutils.generate_uuid(),
                'sha256': blob_meta.get("sha256"), 'status': 'active',
                'external': location_type == 'external', 'content_type': None}

        lock_key = "%s:%s" % (type_name, artifact_id)
        with self.lock_engine.acquire(context, lock_key):
            af = self._show_artifact(context, type_name, artifact_id)
            policy.authorize(action_name, af.to_dict(), context)
            if self._get_blob_info(af, field_name, blob_key):
                msg = _("Blob %(blob)s already exists for artifact "
                        "%(af)s") % {'blob': field_name, 'af': af.id}
                raise exception.Conflict(message=msg)
            utils.validate_change_allowed(af, field_name)
            af.pre_add_location_hook(
                context, af, field_name, location, blob_key)
            af = self._save_blob_info(context, af, field_name, blob_key, blob)

        LOG.info("External location %(location)s has been created "
                 "successfully for artifact %(artifact)s blob %(blob)s",
                 {'location': location, 'artifact': af.id,
                  'blob': blob_name})

        af.post_add_location_hook(context, af, field_name, blob_key)
        Notifier.notify(context, action_name, af)
        return af.to_dict()

    def _calculate_allowed_space(self, context, af, field_name,
                                 content_length=None, blob_key=None):
        """Calculate the maximum amount of data user can upload to the blob."""
        # As a default we take the maximum blob size
        blob_name = self._generate_blob_name(field_name, blob_key)

        max_blob_size = af.get_max_blob_size(field_name)

        if blob_key is not None:
            # For folders we also compare it with the maximum folder size
            blobs_dict = getattr(af, field_name)
            overall_folder_size = sum(
                blob["size"] for blob in blobs_dict.values()
                if blob["size"] is not None)
            available_folder_space = af.get_max_folder_size(
                field_name) - overall_folder_size  # always non-negative
            max_blob_size = min(max_blob_size, available_folder_space)

        # check quotas
        quota_size = quota.verify_uploaded_data_amount(
            context, af.get_type_name(), content_length)

        if content_length is None:
            # if no content_length was provided we have to allocate
            # all allowed space for the blob. It's minimum of max blob size
            # and available quota limit. -1 means that user don't have upload
            # limits.
            size = max_blob_size if quota_size == -1 else min(
                max_blob_size, quota_size)
        else:
            if content_length > max_blob_size:
                msg = _("Can't upload %(content_length)d bytes of data to "
                        "blob %(blob_name)s. Its max allowed size is "
                        "%(max_blob_size)d") % {
                    'content_length': content_length,
                    'blob_name': blob_name,
                    'max_blob_size': max_blob_size}
                raise exception.RequestEntityTooLarge(msg)
            size = content_length

        return size

    def upload_blob(self, context, type_name, artifact_id, field_name, fd,
                    content_type, content_length=None, blob_key=None):
        """Upload Artifact blob.

        :param context: user context
        :param type_name: name of artifact type
        :param artifact_id: id of the artifact to be updated
        :param field_name: name of blob or blob dict field
        :param fd: file descriptor that Glare uses to upload the file
        :param content_type: data content-type
        :param content_length: amount of data user wants to upload
        :param blob_key: if field_name is blob dict it specifies key
         in this dictionary
        :return: dict representation of updated artifact
        """
        blob_name = self._generate_blob_name(field_name, blob_key)
        blob_id = uuidutils.generate_uuid()
        blob_info = {'url': None, 'size': None, 'md5': None, 'sha1': None,
                     'sha256': None, 'id': blob_id, 'status': 'saving',
                     'external': False, 'content_type': content_type}

        # Step 1. Initialize blob
        lock_key = "%s:%s" % (type_name, artifact_id)
        with self.lock_engine.acquire(context, lock_key):
            af = self._show_artifact(context, type_name, artifact_id)
            action_name = "artifact:upload"
            policy.authorize(action_name, af.to_dict(), context)

            # create an an empty blob instance in db with 'saving' status
            if self._get_blob_info(af, field_name, blob_key):
                msg = _("Blob %(blob)s already exists for artifact "
                        "%(af)s") % {'blob': field_name, 'af': af.id}
                raise exception.Conflict(message=msg)
            utils.validate_change_allowed(af, field_name)
            blob_info['size'] = self._calculate_allowed_space(
                context, af, field_name, content_length, blob_key)

            af = self._save_blob_info(
                context, af, field_name, blob_key, blob_info)

        LOG.debug("Parameters validation for artifact %(artifact)s blob "
                  "upload passed for blob %(blob_name)s. "
                  "Start blob uploading to backend.",
                  {'artifact': af.id, 'blob_name': blob_name})

        # Step 2. Call pre_upload_hook and upload data to the store
        try:
            try:
                # call upload hook first
                if hasattr(af, 'validate_upload'):
                    LOG.warning("Method 'validate_upload' was deprecated. "
                                "Please use 'pre_upload_hook' instead.")
                    fd, path = tpool.execute(
                        af.validate_upload, context, af, field_name, fd)
                else:
                    fd = tpool.execute(af.pre_upload_hook,
                                       context, af, field_name, blob_key, fd)
            except exception.GlareException:
                raise
            except Exception as e:
                raise exception.BadRequest(message=str(e))

            default_store = getattr(
                CONF, 'artifact_type:' + type_name).default_store
            # use global parameter if default store isn't set per artifact type
            if default_store is None:
                default_store = CONF.glance_store.default_store

            location_uri, size, checksums = store_api.save_blob_to_store(
                blob_id, fd, context, blob_info['size'],
                store_type=default_store)
            blob_info.update({'url': location_uri,
                              'status': 'active',
                              'size': size})
            blob_info.update(checksums)
        except Exception:
            # if upload failed remove blob from db and storage
            with excutils.save_and_reraise_exception(logger=LOG):
                self._save_blob_info(
                    context, af, field_name, blob_key, None)

        LOG.info("Successfully finished blob uploading for artifact "
                 "%(artifact)s blob field %(blob)s.",
                 {'artifact': af.id, 'blob': blob_name})

        # Step 3. Change blob status to 'active'
        with self.lock_engine.acquire(context, lock_key):
            af = af.show(context, artifact_id)
            af = self._save_blob_info(
                context, af, field_name, blob_key, blob_info)

        af.post_upload_hook(context, af, field_name, blob_key)

        Notifier.notify(context, action_name, af)
        return af.to_dict()

    def download_blob(self, context, type_name, artifact_id, field_name,
                      blob_key=None):
        """Download binary data from Glare Artifact.

        :param context: user context
        :param type_name: name of artifact type
        :param artifact_id: id of the artifact to be updated
        :param field_name: name of blob or blob dict field
        :param blob_key: if field_name is blob dict it specifies key
         in this dict
        :return: file iterator for requested file
        """
        af = self._show_artifact(context, type_name, artifact_id,
                                 read_only=True)
        policy.authorize("artifact:download", af.to_dict(), context)

        blob_name = self._generate_blob_name(field_name, blob_key)

        if af.status == 'deleted':
            msg = _("Cannot download data when artifact is deleted")
            raise exception.Forbidden(message=msg)

        blob = self._get_blob_info(af, field_name, blob_key)
        if blob is None:
            msg = _("No data found for blob %s") % blob_name
            raise exception.NotFound(message=msg)
        if blob['status'] != 'active':
            msg = _("%s is not ready for download") % blob_name
            raise exception.Conflict(message=msg)

        af.pre_download_hook(context, af, field_name, blob_key)

        meta = {'md5': blob.get('md5'),
                'sha1': blob.get('sha1'),
                'sha256': blob.get('sha256'),
                'external': blob.get('external')}
        if blob['external']:
            data = {'url': blob['url']}
        else:
            data = store_api.load_from_store(uri=blob['url'], context=context)
            meta['size'] = blob.get('size')
            meta['content_type'] = blob.get('content_type')

        try:
            # call download hook in the end
            data = af.post_download_hook(
                context, af, field_name, blob_key, data)
        except exception.GlareException:
            raise
        except Exception as e:
            raise exception.BadRequest(message=str(e))

        return data, meta

    def delete_external_blob(self, context, type_name, artifact_id,
                             field_name, blob_key=None):
        """Delete artifact blob with external location.

        :param context: user context
        :param type_name: name of artifact type
        :param artifact_id: id of artifact with the blob to delete
        :param field_name: name of blob or blob dict field
        :param blob_key: if field_name is blob dict it specifies key
         in this dictionary
        """
        af = self._show_artifact(context, type_name, artifact_id)
        action_name = 'artifact:delete_blob'
        policy.authorize(action_name, af.to_dict(), context)

        blob_name = self._generate_blob_name(field_name, blob_key)

        blob = self._get_blob_info(af, field_name, blob_key)
        if blob is None:
            msg = _("Blob %s wasn't found for artifact") % blob_name
            raise exception.NotFound(message=msg)
        if not blob['external']:
            msg = _("Blob %s is not external") % blob_name
            raise exception.Forbidden(message=msg)

        af = self._save_blob_info(context, af, field_name, blob_key, None)

        Notifier.notify(context, action_name, af)
        return af.to_dict()

    @staticmethod
    def set_quotas(context, values):
        """Set quota records in Glare.

        :param context: user request context
        :param values: dict with quota values to set
        """
        action_name = "artifact:set_quotas"
        policy.authorize(action_name, {}, context)
        qs = quota.set_quotas(values)
        Notifier.notify(context, action_name, qs)

    def list_all_quotas(self, context):
        """Get detailed info about all available quotas.

        :param context: user request context
        :return: dict with definitions of redefined quotas for all projects
         and global defaults
        """
        action_name = "artifact:list_all_quotas"
        policy.authorize(action_name, {}, context)
        return {
            'quotas': quota.list_quotas(),
            'global_quotas': self.config_quotas
        }

    def list_project_quotas(self, context, project_id=None):
        """Get detailed info about project quotas.

        :param context: user request context
        :param project_id: id of the project for which to show quotas
        :return: definition of requested quotas for the project
        """
        project_id = project_id or context.tenant
        action_name = "artifact:list_project_quotas"
        policy.authorize(action_name, {'project_id': project_id}, context)
        qs = self.config_quotas.copy()
        qs.update(quota.list_quotas(project_id)[project_id])
        return {project_id: qs}



# Copyright 2017 - Nokia Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import mock

from glare.common import exception as exc
from glare.db import artifact_api
from glare.tests.unit import base


class TestLocations(base.BaseTestArtifactAPI):
    """Test adding custom locations."""

    def setUp(self):
        super(TestLocations, self).setUp()
        values = {'name': 'ttt', 'version': '1.0'}
        self.sample_artifact = self.controller.create(
            self.req, 'sample_artifact', values)
        self.ct = 'application/vnd+openstack.glare-custom-location+json'

    def test_add_location(self):
        with mock.patch(
                'glare.common.store_api.save_blob_to_store') as mocked_add:
            body = {'url': 'https://FAKE_LOCATION.com',
                    'md5': "fake", 'sha1': "fake_sha", "sha256": "fake_sha256"}
            self.controller.upload_blob(
                self.req, 'sample_artifact', self.sample_artifact['id'],
                'blob', body, self.ct)
            art = self.controller.show(self.req, 'sample_artifact',
                                       self.sample_artifact['id'])
            self.assertEqual('active', art['blob']['status'])
            self.assertEqual('fake', art['blob']['md5'])
            self.assertEqual('fake_sha', art['blob']['sha1'])
            self.assertEqual('fake_sha256', art['blob']['sha256'])
            self.assertIsNone(art['blob']['size'])
            self.assertIsNone(art['blob']['content_type'])
            self.assertEqual('https://FAKE_LOCATION.com',
                             art['blob']['url'])
            self.assertNotIn('id', art['blob'])
            self.assertEqual(0, mocked_add.call_count)

        # Adding location for the second time leads to Conflict error
        body = {'url': 'https://ANOTHER_FAKE_LOCATION.com',
                'md5': "fake", 'sha1': "fake_sha", "sha256": "fake_sha256"}
        self.assertRaises(
            exc.Conflict, self.controller.upload_blob,
            self.req, 'sample_artifact', self.sample_artifact['id'],
            'blob', body, self.ct)

    def test_add_dict_location(self):
        with mock.patch(
                'glare.common.store_api.save_blob_to_store') as mocked_add:
            body = {'url': 'https://FAKE_LOCATION.com',
                    'md5': "fake", 'sha1': "fake_sha", "sha256": "fake_sha256"}
            self.controller.upload_blob(
                self.req, 'sample_artifact', self.sample_artifact['id'],
                'dict_of_blobs/blob', body, self.ct)
            art = self.controller.show(self.req, 'sample_artifact',
                                       self.sample_artifact['id'])
            self.assertEqual('active', art['dict_of_blobs']['blob']['status'])
            self.assertEqual('fake', art['dict_of_blobs']['blob']['md5'])
            self.assertEqual('fake_sha', art['dict_of_blobs']['blob']['sha1'])
            self.assertEqual('fake_sha256',
                             art['dict_of_blobs']['blob']['sha256'])
            self.assertIsNone(art['dict_of_blobs']['blob']['size'])
            self.assertIsNone(art['dict_of_blobs']['blob']['content_type'])
            self.assertEqual('https://FAKE_LOCATION.com',
                             art['dict_of_blobs']['blob']['url'])
            self.assertNotIn('id', art['blob'])
            self.assertEqual(0, mocked_add.call_count)

        # Adding location for the second time leads to Conflict error
        body = {'url': 'https://ANOTHER_FAKE_LOCATION.com',
                'md5': "fake", 'sha1': "fake_sha", "sha256": "fake_sha256"}
        self.assertRaises(
            exc.Conflict, self.controller.upload_blob,
            self.req, 'sample_artifact', self.sample_artifact['id'],
            'dict_of_blobs/blob', body, self.ct)

    def test_add_location_saving_blob(self):
        body = {'url': 'https://FAKE_LOCATION.com',
                'md5': "fake", 'sha1': "fake_sha", "sha256": "fake_sha256"}
        self.controller.upload_blob(
            self.req, 'sample_artifact', self.sample_artifact['id'],
            'blob', body, self.ct)
        art = self.controller.show(self.req, 'sample_artifact',
                                   self.sample_artifact['id'])

        # Change status of the blob to 'saving'
        art['blob']['status'] = 'saving'
        artifact_api.ArtifactAPI().update_blob(
            self.req.context, self.sample_artifact['id'],
            {'blob': art['blob']})
        art = self.controller.show(
            self.req, 'sample_artifact', self.sample_artifact['id'])
        self.assertEqual('saving', art['blob']['status'])

        body = {'url': 'https://ANOTHER_FAKE_LOCATION.com',
                'md5': "fake", 'sha1': "fake_sha", "sha256": "fake_sha256"}
        self.assertRaises(
            exc.Conflict, self.controller.upload_blob,
            self.req, 'sample_artifact', self.sample_artifact['id'],
            'blob', body, self.ct)

    def test_too_long_location_url(self):
        body = {'url': 'http://FAKE_LOCATION%s.com' % ('a' * 2049),
                'md5': "fake", 'sha1': "fake_sha", "sha256": "fake_sha256"}
        self.assertRaises(
            exc.BadRequest, self.controller.upload_blob,
            self.req, 'sample_artifact', self.sample_artifact['id'],
            'blob', body, self.ct)

    def test_internal_location(self):
        # allow regular user to set internal locations
        rule = {"artifact:set_internal_location": "rule:admin_or_owner"}
        self.policy(rule)

        art_id = self.sample_artifact['id']

        # Setting locations with forbidden schemas fails
        forbidden_schemes = ('file', 'filesystem', 'swift+config', 'sql')
        for scheme in forbidden_schemes:
            body = {'md5': 'fake', 'sha1': 'fake_sha', 'sha256': 'fake_sha256',
                    'location_type': 'internal',
                    'url': scheme + '://FAKE_LOCATION.com'}
            self.assertRaises(
                exc.Forbidden, self.controller.upload_blob,
                self.req, 'sample_artifact', art_id, 'blob', body, self.ct)

        # Setting locations with unknown schemes fail
        body = {'md5': 'fake', 'sha1': 'fake_sha', 'sha256': 'fake_sha256',
                'location_type': 'internal',
                'url': 'UNKNOWN://FAKE_LOCATION.com'}
        self.assertRaises(
            exc.BadRequest, self.controller.upload_blob,
            self.req, 'sample_artifact', art_id, 'blob', body, self.ct)

        with mock.patch(
                'glare.common.store_api.save_blob_to_store') as mocked_add:
            body = {'md5': 'fake', 'sha1': 'fake_sha', 'sha256': 'fake_sha256',
                    'location_type': 'internal',
                    'url': 'https://FAKE_LOCATION.com'}
            self.controller.upload_blob(
                self.req, 'sample_artifact', art_id, 'blob', body, self.ct)
            art = self.controller.show(self.req, 'sample_artifact', art_id)

            self.assertFalse(art['blob']['external'])
            self.assertEqual('active', art['blob']['status'])
            self.assertEqual('fake', art['blob']['md5'])
            self.assertEqual('fake_sha', art['blob']['sha1'])
            self.assertEqual('fake_sha256', art['blob']['sha256'])
            self.assertIsNone(art['blob']['size'])
            self.assertIsNone(art['blob']['content_type'])
            self.assertEqual('/artifacts/sample_artifact/%s/blob' % art_id,
                             art['blob']['url'])
            self.assertNotIn('id', art['blob'])
            self.assertEqual(0, mocked_add.call_count)

        # deletion of artifact leads to the deletion of data under the internal
        # location
        with mock.patch('glare.common.store_api.delete_blob') as mocked_del:
            self.controller.delete(self.req, 'sample_artifact', art_id)
            self.assertEqual(1, mocked_del.call_count)



# Copyright 2017 - Nokia Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from six import BytesIO
from uuid import uuid4

from glare.common import exception as exc
from glare.db import artifact_api
from glare.tests.unit import base


class TestArtifactUpdate(base.BaseTestArtifactAPI):

    """Test Glare artifact updates."""

    def setUp(self):
        super(TestArtifactUpdate, self).setUp()
        values = {'name': 'ttt', 'version': '1.0'}
        self.sample_artifact = self.controller.create(
            self.req, 'sample_artifact', values)

    def test_basic_update(self):
        changes = [
            {'op': 'replace', 'path': '/name', 'value': 'new_name'},
            {'op': 'replace', 'path': '/version', 'value': '1.0.0'},
            {'op': 'replace', 'path': '/description', 'value': 'Test'},
            {'op': 'replace', 'path': '/tags', 'value': ['tag1', 'tag2']},
            {'op': 'replace', 'path': '/metadata', 'value': {'k': 'v'}},
        ]
        res = self.update_with_values(changes)
        self.assertEqual('new_name', res['name'])
        self.assertEqual('1.0.0', res['version'])
        self.assertEqual('Test', res['description'])
        self.assertEqual({'tag1', 'tag2'}, set(res['tags']))
        self.assertEqual({'k': 'v'}, res['metadata'])

    def test_update_replace_values(self):
        changes = [
            {'op': 'replace', 'path': '/int1', 'value': 1},
            {'op': 'replace', 'path': '/float1', 'value': 1.0},
            {'op': 'replace', 'path': '/str1', 'value': 'Test'},
            {'op': 'replace', 'path': '/list_of_int', 'value': [0, 1]},
            {'op': 'replace', 'path': '/dict_of_str', 'value': {'k': 'v'}},
        ]
        res = self.update_with_values(changes)
        self.assertEqual(1, res['int1'])
        self.assertEqual(1.0, res['float1'])
        self.assertEqual('Test', res['str1'])
        self.assertEqual([0, 1], res['list_of_int'])
        self.assertEqual({'k': 'v'}, res['dict_of_str'])

        changes = [
            {'op': 'replace', 'path': '/int1', 'value': 2},
            {'op': 'replace', 'path': '/float1', 'value': 2.0},
            {'op': 'replace', 'path': '/str1', 'value': 'New_Test'},
            {'op': 'replace', 'path': '/list_of_int/1', 'value': 4},
            {'op': 'replace', 'path': '/dict_of_str/k', 'value': 'new_val'},
        ]
        res = self.update_with_values(changes)
        self.assertEqual(2, res['int1'])
        self.assertEqual(2.0, res['float1'])
        self.assertEqual('New_Test', res['str1'])
        self.assertEqual([0, 4], res['list_of_int'])
        self.assertEqual({'k': 'new_val'}, res['dict_of_str'])

    def test_update_no_artifact_type(self):
        changes = [{'op': 'replace', 'path': '/name', 'value': 'new_name'}]
        self.update_with_values(
            changes, exc_class=exc.NotFound, art_type='wrong_type')

    def test_update_name_version(self):
        # Create additional artifacts
        values = {'name': 'ttt', 'version': '2.0'}
        self.controller.create(self.req, 'sample_artifact', values)
        values = {'name': 'ddd', 'version': '1.0'}
        self.controller.create(self.req, 'sample_artifact', values)

        # This name/version is already taken
        changes = [{'op': 'replace', 'path': '/version',
                    'value': '2.0'}]
        self.assertRaises(exc.Conflict, self.update_with_values, changes)
        changes = [{'op': 'replace', 'path': '/name',
                    'value': 'ddd'}]
        self.assertRaises(exc.Conflict, self.update_with_values, changes)

        # Test coercing
        # name
        changes = [{'op': 'replace', 'path': '/name', 'value': True}]
        res = self.update_with_values(changes)
        self.assertEqual('True', res['name'])

        changes = [{'op': 'replace', 'path': '/name', 'value': 1.0}]
        res = self.update_with_values(changes)
        self.assertEqual('1.0', res['name'])

        changes = [{'op': 'replace', 'path': '/name', 'value': "tt:t"}]
        res = self.update_with_values(changes)
        self.assertEqual('tt:t', res['name'])

        # version
        changes = [{'op': 'replace', 'path': '/version', 'value': 2.0}]
        res = self.update_with_values(changes)
        self.assertEqual('2.0.0', res['version'])

        changes = [{'op': 'replace', 'path': '/version', 'value': '1-alpha'}]
        res = self.update_with_values(changes)
        self.assertEqual('1.0.0-alpha', res['version'])

        changes = [{'op': 'replace', 'path': '/version', 'value': '1:0'}]
        res = self.update_with_values(changes)
        self.assertEqual('1.0.0-0', res['version'])

    def test_update_deleted_artifact(self):
        # Enable delayed delete
        self.config(delayed_delete=True)
        # Delete artifact and check its status
        self.controller.delete(self.req, 'sample_artifact',
                               self.sample_artifact['id'])
        art = self.controller.show(self.req, 'sample_artifact',
                                   self.sample_artifact['id'])
        self.assertEqual('deleted', art['status'])

        changes = [{'op': 'replace', 'path': '/int1',
                    'value': 1}]
        self.assertRaises(exc.Forbidden, self.update_with_values, changes)

        changes = [{'op': 'replace', 'path': '/name',
                    'value': 'new'}]
        self.assertRaises(exc.Forbidden, self.update_with_values, changes)

    def test_update_lists(self):
        changes = [{'op': 'replace', 'path': '/list_of_str',
                    'value': ['val1', 'val2']}]
        res = self.update_with_values(changes)
        self.assertEqual({'val1', 'val2'}, set(res['list_of_str']))

        changes = [{'op': 'remove', 'path': '/list_of_str/0'}]
        res = self.update_with_values(changes)
        self.assertEqual(['val2'], res['list_of_str'])

        changes = [{'op': 'replace', 'path': '/list_of_str', 'value': None}]
        res = self.update_with_values(changes)
        self.assertEqual([], res['list_of_str'])

        changes = [{'op': 'add', 'path': '/list_of_str/-', 'value': 'val1'}]
        res = self.update_with_values(changes)
        self.assertEqual(['val1'], res['list_of_str'])

        changes = [{'op': 'replace', 'path': '/list_of_str/0',
                    'value': 'val2'}]
        res = self.update_with_values(changes)
        self.assertEqual(['val2'], res['list_of_str'])

        changes = [{'op': 'replace', 'path': '/list_of_str', 'value': []}]
        res = self.update_with_values(changes)
        self.assertEqual([], res['list_of_str'])

        changes = [{'op': 'replace', 'path': '/list_of_str', 'value': {}}]
        self.update_with_values(changes, exc_class=exc.BadRequest)

        changes = [{'op': 'replace', 'path': '/list_of_str',
                    'value': {'a': 'b'}}]
        self.update_with_values(changes, exc_class=exc.BadRequest)

        changes = [{'op': 'replace', 'path': '/list_of_str',
                    'value': [['a']]}]
        self.update_with_values(changes, exc_class=exc.BadRequest)

        changes = [{'op': 'remove', 'path': '/list_of_str/-',
                    'value': 'val3'}]
        self.update_with_values(changes, exc_class=exc.BadRequest)

    def test_update_dicts(self):
        changes = [{'op': 'replace', 'path': '/dict_of_str',
                    'value': {'k1': 'v1', 'k2': 'v2'}}]
        res = self.update_with_values(changes)
        self.assertEqual({'k1': 'v1', 'k2': 'v2'}, res['dict_of_str'])

        changes = [{'op': 'remove', 'path': '/dict_of_str/k1'}]
        res = self.update_with_values(changes)
        self.assertEqual({'k2': 'v2'}, res['dict_of_str'])

        changes = [{'op': 'replace', 'path': '/dict_of_str', 'value': None}]
        res = self.update_with_values(changes)
        self.assertEqual({}, res['dict_of_str'])

        changes = [{'op': 'add', 'path': '/dict_of_str/k1', 'value': 'v1'}]
        res = self.update_with_values(changes)
        self.assertEqual({'k1': 'v1'}, res['dict_of_str'])

        changes = [{'op': 'replace', 'path': '/dict_of_str/k1',
                    'value': 'v2'}]
        res = self.update_with_values(changes)
        self.assertEqual({'k1': 'v2'}, res['dict_of_str'])

        changes = [{'op': 'replace', 'path': '/dict_of_str', 'value': {}}]
        res = self.update_with_values(changes)
        self.assertEqual({}, res['dict_of_str'])

        changes = [{'op': 'replace', 'path': '/dict_of_str', 'value': []}]
        self.update_with_values(changes, exc_class=exc.BadRequest)

        changes = [{'op': 'replace', 'path': '/dict_of_str',
                    'value': ['a']}]
        self.update_with_values(changes, exc_class=exc.BadRequest)

        changes = [{'op': 'replace', 'path': '/dict_of_str/k10',
                    'value': {'k100': 'v100'}}]
        self.update_with_values(changes, exc_class=exc.BadRequest)

    def test_update_artifact_wrong_parameters(self):
        changes = [{'op': 'replace', 'path': '/name', 'value': ''}]
        self.update_with_values(changes, exc_class=exc.BadRequest)

        changes = [{'op': 'replace', 'path': '/name', 'value': 'a' * 256}]
        self.update_with_values(changes, exc_class=exc.BadRequest)

        changes = [{'op': 'replace', 'path': '/version', 'value': ''}]
        self.update_with_values(changes, exc_class=exc.BadRequest)

        changes = [{'op': 'replace', 'path': '/version', 'value': 'invalid'}]
        self.update_with_values(changes, exc_class=exc.BadRequest)

        changes = [{'op': 'replace', 'path': '/version', 'value': -1}]
        self.update_with_values(changes, exc_class=exc.BadRequest)

        changes = [{'op': 'replace', 'path': '/description',
                    'value': 'a' * 4097}]
        self.update_with_values(changes, exc_class=exc.BadRequest)

        changes = [{'op': 'replace', 'path': '/tags', 'value': ['a' * 256]}]
        self.update_with_values(changes, exc_class=exc.BadRequest)

        changes = [{'op': 'replace', 'path': '/tags', 'value': ['']}]
        self.update_with_values(changes, exc_class=exc.BadRequest)

        changes = [{'op': 'replace', 'path': '/tags', 'value': ['a/a']}]
        self.update_with_values(changes, exc_class=exc.BadRequest)

        changes = [{'op': 'replace', 'path': '/tags', 'value': ['a,a']}]
        self.update_with_values(changes, exc_class=exc.BadRequest)

        changes = [{'op': 'replace', 'path': '/tags',
                    'value': [str(i) for i in range(256)]}]
        self.update_with_values(changes, exc_class=exc.BadRequest)

        changes = [{'op': 'replace', 'path': '/metadata',
                    'value': {'key': 'a' * 256}}]
        self.update_with_values(changes, exc_class=exc.BadRequest)

        changes = [{'op': 'replace', 'path': '/metadata',
                    'value': {'': 'a'}}]
        self.update_with_values(changes, exc_class=exc.BadRequest)

        changes = [{'op': 'replace', 'path': '/metadata',
                    'value': {'a' * 256: 'a'}}]
        self.update_with_values(changes, exc_class=exc.BadRequest)

        changes = [{'op': 'replace', 'path': '/metadata',
                    'value': {('a' + str(i)): 'a' for i in range(256)}}]
        self.update_with_values(changes, exc_class=exc.BadRequest)

        changes = [{'op': 'replace', 'path': '/int1', 'value': 'aaa'}]
        self.update_with_values(changes, exc_class=exc.BadRequest)

        changes = [{'op': 'replace', 'path': '/float1', 'value': 'aaa'}]
        self.update_with_values(changes, exc_class=exc.BadRequest)

    def test_update_artifact_not_existing_field(self):
        changes = [{'op': 'replace', 'path': '/wrong_field', 'value': 'a'}]
        self.update_with_values(changes, exc_class=exc.BadRequest)

        changes = [{'op': 'replace', 'path': '/', 'value': 'a'}]
        self.update_with_values(changes, exc_class=exc.BadRequest)

        changes = [{'op': 'add', 'path': '/wrong_field', 'value': 'a'}]
        self.update_with_values(changes, exc_class=exc.BadRequest)

        changes = [{'op': 'add', 'path': '/', 'value': 'a'}]
        self.update_with_values(changes, exc_class=exc.BadRequest)

    def test_update_artifact_remove_field(self):
        changes = [{'op': 'remove', 'path': '/name'}]
        self.update_with_values(changes, exc_class=exc.BadRequest)

        changes = [{'op': 'remove', 'path': '/list_of_int/10'}]
        self.update_with_values(changes, exc_class=exc.BadRequest)

        changes = [{'op': 'remove', 'path': '/status'}]
        self.update_with_values(changes, exc_class=exc.BadRequest)

        changes = [
            {'op': 'add', 'path': '/list_of_int/-', 'value': 4},
            {'op': 'add', 'path': '/dict_of_str/k', 'value': 'new_val'},
        ]
        self.update_with_values(changes)
        changes = [{'op': 'remove', 'path': '/list_of_int/0'}]
        res = self.update_with_values(changes)
        self.assertEqual([], res['list_of_int'])
        changes = [{'op': 'remove', 'path': '/dict_of_str/k'}]
        res = self.update_with_values(changes)
        self.assertEqual({}, res['dict_of_str'])

    def test_update_artifact_blob(self):
        changes = [{'op': 'replace', 'path': '/blob', 'value': 'a'}]
        self.update_with_values(changes, exc_class=exc.BadRequest)

    def test_update_artifact_system_fields(self):
        changes = [{'op': 'replace', 'path': '/id',
                    'value': '5fdeba9a-ba12-4147-bb8a-a8daada84222'}]
        self.update_with_values(changes, exc_class=exc.Forbidden)

        changes = [{'op': 'replace', 'path': '/created_at',
                    'value': '2000-01-01'}]
        self.update_with_values(changes, exc_class=exc.Forbidden)

        changes = [{'op': 'replace', 'path': '/updated_at',
                    'value': '2000-01-01'}]
        self.update_with_values(changes, exc_class=exc.Forbidden)

        changes = [{'op': 'replace', 'path': '/activated_at',
                    'value': '2000-01-01'}]
        self.update_with_values(changes, exc_class=exc.Forbidden)

        changes = [{'op': 'replace', 'path': '/owner', 'value': 'new_owner'}]
        self.update_with_values(changes, exc_class=exc.Forbidden)

        changes = [{'op': 'replace', 'path': '/system_attribute',
                    'value': 'some_value'}]
        self.update_with_values(changes, exc_class=exc.Forbidden)

    def test_update_artifact_visibility(self):
        self.req = self.get_fake_request(user=self.users['admin'])

        changes = [{'op': 'replace', 'path': '/visibility',
                    'value': 'wrong_value'}]
        self.update_with_values(changes, exc_class=exc.BadRequest)

        changes = [{'op': 'replace', 'path': '/visibility',
                    'value': 'public'}]
        self.update_with_values(changes, exc_class=exc.Forbidden)

        changes = [{'op': 'replace', 'path': '/visibility',
                    'value': None}]
        self.update_with_values(changes, exc_class=exc.BadRequest)

        changes = [{'op': 'replace', 'path': '/string_required',
                    'value': 'some_string'},
                   {'op': 'replace', 'path': '/status',
                    'value': 'active'}]
        res = self.update_with_values(changes)
        self.assertEqual('active', res['status'])
        self.assertEqual('some_string', res['string_required'])

        changes = [{'op': 'replace', 'path': '/visibility', 'value': 'public'}]
        res = self.update_with_values(changes)
        self.assertEqual('public', res['visibility'])

        changes = [{'op': 'replace', 'path': '/visibility', 'value': 'public'}]
        res = self.update_with_values(changes)
        self.assertEqual('public', res['visibility'])

        changes = [{'op': 'replace', 'path': '/visibility',
                    'value': 'private'}]
        self.update_with_values(changes, exc_class=exc.Forbidden)

    def test_update_artifact_status(self):
        self.req = self.get_fake_request(user=self.users['admin'])

        changes = [{'op': 'replace', 'path': '/status',
                    'value': 'wrong_value'}]
        self.update_with_values(changes, exc_class=exc.BadRequest)

        # It's forbidden to activate artifact until required_on_activate field
        # 'string_required' is set
        changes = [{'op': 'replace', 'path': '/status',
                    'value': 'active'}]
        self.update_with_values(changes, exc_class=exc.Forbidden)

        changes = [{'op': 'replace', 'path': '/status',
                    'value': None}]
        self.update_with_values(changes, exc_class=exc.BadRequest)

        # It's forbidden to deactivate drafted artifact
        changes = [{'op': 'replace', 'path': '/status',
                    'value': 'deactivated'}]
        self.update_with_values(changes, exc_class=exc.Forbidden)

        changes = [{'op': 'replace', 'path': '/string_required',
                    'value': 'some_string'}]
        res = self.update_with_values(changes)
        self.assertEqual('some_string', res['string_required'])

        # It's impossible to activate the artifact when it has 'saving' blobs
        self.controller.upload_blob(
            self.req, 'sample_artifact', self.sample_artifact['id'], 'blob',
            BytesIO(b'aaa'), 'application/octet-stream')
        self.sample_artifact = self.controller.show(
            self.req, 'sample_artifact', self.sample_artifact['id'])

        # Change status of the blob to 'saving'
        self.sample_artifact['blob']['status'] = 'saving'
        artifact_api.ArtifactAPI().update_blob(
            self.req.context, self.sample_artifact['id'],
            {'blob': self.sample_artifact['blob']})
        self.sample_artifact = self.controller.show(
            self.req, 'sample_artifact', self.sample_artifact['id'])
        self.assertEqual('saving', self.sample_artifact['blob']['status'])

        # Now activating of the artifact leads to Conflict
        changes = [{'op': 'replace', 'path': '/status', 'value': 'active'}]
        self.assertRaises(exc.Conflict, self.update_with_values, changes)

        # Reverting status of the blob to active again
        self.sample_artifact['blob']['status'] = 'active'
        artifact_api.ArtifactAPI().update_blob(
            self.req.context, self.sample_artifact['id'],
            {'blob': self.sample_artifact['blob']})
        self.sample_artifact = self.controller.show(
            self.req, 'sample_artifact', self.sample_artifact['id'])
        self.assertEqual('active', self.sample_artifact['blob']['status'])

        # It's possible to change artifact status with other fields in
        # one request
        changes = [
            {'op': 'replace', 'path': '/name', 'value': 'new_name'},
            {'op': 'replace', 'path': '/status', 'value': 'active'}
        ]
        self.sample_artifact = self.update_with_values(changes)
        self.assertEqual('new_name', self.sample_artifact['name'])
        self.assertEqual('active', self.sample_artifact['status'])

        changes = [{'op': 'replace', 'path': '/status', 'value': 'active'}]
        res = self.update_with_values(changes)
        self.assertEqual('active', res['status'])

        # It's possible to change artifact status with other fields in
        # one request
        changes = [
            {'op': 'replace', 'path': '/string_mutable', 'value': 'str'},
            {'op': 'replace', 'path': '/status', 'value': 'deactivated'}
        ]
        self.sample_artifact = self.update_with_values(changes)
        self.assertEqual('str', self.sample_artifact['string_mutable'])
        self.assertEqual('deactivated', self.sample_artifact['status'])

        changes = [{'op': 'replace', 'path': '/status',
                    'value': 'deactivated'}]
        res = self.update_with_values(changes)
        self.assertEqual('deactivated', res['status'])

        # It's possible to change artifact status with other fields in
        # one request
        changes = [
            {'op': 'replace', 'path': '/status', 'value': 'active'},
            {'op': 'replace', 'path': '/description', 'value': 'test'},
        ]
        self.sample_artifact = self.update_with_values(changes)
        self.assertEqual('test', self.sample_artifact['description'])
        self.assertEqual('active', self.sample_artifact['status'])

        changes = [{'op': 'replace', 'path': '/status', 'value': 'active'}]
        res = self.update_with_values(changes)
        self.assertEqual('active', res['status'])

        changes = [{'op': 'replace', 'path': '/status',
                    'value': None}]
        self.update_with_values(changes, exc_class=exc.BadRequest)

        # Enable delayed delete
        self.config(delayed_delete=True)
        # Delete artifact and check its status
        self.controller.delete(self.req, 'sample_artifact',
                               self.sample_artifact['id'])
        art = self.controller.show(self.req, 'sample_artifact',
                                   self.sample_artifact['id'])
        self.assertEqual('deleted', art['status'])

        changes = [{'op': 'replace', 'path': '/status', 'value': 'active'}]
        self.assertRaises(exc.Forbidden,
                          self.update_with_values, changes)

    def test_update_artifact_mutable_fields(self):
        changes = [{'op': 'replace', 'path': '/string_required',
                    'value': 'some_string'}]
        res = self.update_with_values(changes)
        self.assertEqual('some_string', res['string_required'])

        changes = [{'op': 'replace', 'path': '/status', 'value': 'active'}]
        res = self.update_with_values(changes)
        self.assertEqual('active', res['status'])

        changes = [{'op': 'replace', 'path': '/name', 'value': 'new_name'}]
        self.update_with_values(changes, exc_class=exc.Forbidden)

        changes = [{'op': 'replace', 'path': '/metadata', 'value': {'k': 'v'}}]
        self.update_with_values(changes, exc_class=exc.Forbidden)

        changes = [{'op': 'add', 'path': '/metadata/k', 'value': 'v'}]
        self.update_with_values(changes, exc_class=exc.Forbidden)

        changes = [{'op': 'replace', 'path': '/tags', 'value': ['a']}]
        res = self.update_with_values(changes)
        self.assertEqual(['a'], res['tags'])

        changes = [{'op': 'add', 'path': '/tags/-', 'value': 'b'}]
        res = self.update_with_values(changes)
        self.assertEqual({'a', 'b'}, set(res['tags']))

        changes = [{'op': 'replace', 'path': '/description', 'value': 'Test'}]
        res = self.update_with_values(changes)
        self.assertEqual('Test', res['description'])

        changes = [{'op': 'replace', 'path': '/string_mutable',
                    'value': 'some_value'}]
        res = self.update_with_values(changes)
        self.assertEqual('some_value', res['string_mutable'])

    def test_update_artifact_unicode(self):
        name = u'\u0442\u0435\u0441\u0442'
        description = u'\u041E\u043F\u0438\u0441\u0430\u043D\u0438\u0435'
        tags = [u'\u041C\u0435\u0442\u043A\u0430']
        metadata = {'key': u'\u0417\u043D\u0430\u0447\u0435\u043D\u0438\u0435'}
        changes = [
            {'op': 'replace', 'path': '/name', 'value': name},
            {'op': 'replace', 'path': '/version', 'value': '1.0.0'},
            {'op': 'replace', 'path': '/description', 'value': description},
            {'op': 'replace', 'path': '/tags', 'value': tags},
            {'op': 'replace', 'path': '/metadata', 'value': metadata},
        ]
        res = self.update_with_values(changes)

        self.assertEqual(name, res['name'])
        self.assertEqual('1.0.0', res['version'])
        self.assertEqual(self.users['user1']['tenant_id'], res['owner'])
        self.assertEqual('drafted', res['status'])
        self.assertEqual('private', res['visibility'])
        self.assertEqual(description, res['description'])
        self.assertEqual(metadata, res['metadata'])
        self.assertEqual(tags, res['tags'])

    def test_update_artifact_4_byte_unicode(self):
        bad_name = u'A name with forbidden symbol \U0001f62a'
        changes = [
            {'op': 'replace', 'path': '/name', 'value': bad_name}
        ]

        self.assertRaises(exc.BadRequest, self.update_with_values, changes)


class TestLinks(base.BaseTestArtifactAPI):

    """Test Glare artifact link management."""

    def setUp(self):
        super(TestLinks, self).setUp()
        values = {'name': 'ttt', 'version': '1.0'}
        self.sample_artifact = self.controller.create(
            self.req, 'sample_artifact', values)
        values = {'name': 'sss', 'version': '1.0'}
        self.dependency = self.controller.create(
            self.req, 'sample_artifact', values)

    def test_manage_links(self):
        dep_url = "/artifacts/sample_artifact/%s" % self.dependency['id']

        # set valid link
        patch = [{"op": "replace", "path": "/link1", "value": dep_url}]
        res = self.update_with_values(patch)
        self.assertEqual(res['link1'], dep_url)

        # remove link from artifact
        patch = [{"op": "replace", "path": "/link1", "value": None}]
        res = self.update_with_values(patch)
        self.assertIsNone(res['link1'])

        # set invalid external link
        dep_url = "http://example.com/artifacts/" \
                  "sample_artifact/%s" % self.dependency['id']
        patch = [{"op": "replace", "path": "/link1", "value": dep_url}]
        self.assertRaises(exc.BadRequest, self.update_with_values, patch)

        # try to set invalid link
        patch = [{"op": "replace", "path": "/link1", "value": "Invalid"}]
        self.assertRaises(exc.BadRequest, self.update_with_values, patch)

        # try to set link to non-existing artifact
        non_exiting_url = "/artifacts/sample_artifact/%s" % uuid4()
        patch = [{"op": "replace",
                  "path": "/link1",
                  "value": non_exiting_url}]
        self.assertRaises(exc.BadRequest, self.update_with_values, patch)

    def test_manage_dict_of_links(self):
        dep_url = "/artifacts/sample_artifact/%s" % self.dependency['id']

        # set valid link
        patch = [{"op": "add",
                  "path": "/dict_of_links/link1",
                  "value": dep_url}]
        res = self.update_with_values(patch)
        self.assertEqual(res['dict_of_links']['link1'], dep_url)

        # remove link from artifact
        patch = [{"op": "remove",
                  "path": "/dict_of_links/link1"}]
        res = self.update_with_values(patch)
        self.assertNotIn('link1', res['dict_of_links'])

        # set invalid external link
        dep_url = "http://example.com/artifacts/" \
                  "sample_artifact/%s" % self.dependency['id']
        patch = [{"op": "replace",
                  "path": "/dict_of_links/link1", "value": dep_url}]
        self.assertRaises(exc.BadRequest, self.update_with_values, patch)

        # try to set invalid link
        patch = [{"op": "replace",
                  "path": "/dict_of_links/link1",
                  "value": "Invalid"}]
        self.assertRaises(exc.BadRequest, self.update_with_values, patch)

        # try to set link to non-existing artifact
        non_exiting_url = "/artifacts/sample_artifact/%s" % uuid4()
        patch = [{"op": "replace",
                  "path": "/dict_of_links/link1",
                  "value": non_exiting_url}]
        self.assertRaises(exc.BadRequest, self.update_with_values, patch)

    def test_manage_list_of_links(self):
        dep_url = "/artifacts/sample_artifact/%s" % self.dependency['id']

        # set valid link
        patch = [{"op": "add",
                  "path": "/list_of_links/-",
                  "value": dep_url}]
        res = self.update_with_values(patch)
        self.assertEqual(res['list_of_links'][0], dep_url)

        # remove link from artifact
        patch = [{"op": "remove",
                  "path": "/list_of_links/0"}]
        res = self.update_with_values(patch)
        self.assertEqual(0, len(res['list_of_links']))

        # set invalid external link
        dep_url = "http://example.com/artifacts/" \
                  "sample_artifact/%s" % self.dependency['id']
        patch = [{"op": "replace",
                  "path": "/list_of_links/-", "value": dep_url}]
        self.assertRaises(exc.BadRequest, self.update_with_values, patch)

        # try to set invalid link
        patch = [{"op": "add",
                  "path": "/list_of_links/-",
                  "value": "Invalid"}]
        self.assertRaises(exc.BadRequest, self.update_with_values, patch)

        # try to set link to non-existing artifact
        non_exiting_url = "/artifacts/sample_artifact/%s" % uuid4()
        patch = [{"op": "add",
                  "path": "/list_of_links/-",
                  "value": non_exiting_url}]
        self.assertRaises(exc.BadRequest, self.update_with_values, patch)



# Copyright 2017 - Nokia Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import jwt
import memcache
from oslo_config import cfg
from oslo_log import log as logging
from oslo_middleware import base as base_middleware
import pprint
import requests
from six.moves import urllib
import webob.dec

from glare.common import exception
from glare.common import utils
from glare.i18n import _

LOG = logging.getLogger(__name__)

keycloak_oidc_opts = [
    cfg.StrOpt(
        'auth_url',
        default='http://127.0.0.1:8080/auth',
        help='Keycloak base url (e.g. https://my.keycloak:8443/auth)'
    ),
    cfg.StrOpt(
        'user_info_endpoint_url',
        default='/realms/%s/protocol/openid-connect/userinfo',
        help='Endpoint against which authorization will be performed'
    ),
    cfg.StrOpt(
        'certfile',
        help='Required if identity server requires client certificate'
    ),
    cfg.StrOpt(
        'keyfile',
        help='Required if identity server requires client certificate'
    ),
    cfg.StrOpt(
        'cafile',
        help='A PEM encoded Certificate Authority to use when verifying '
             'HTTPs connections. Defaults to system CAs.'
    ),
    cfg.BoolOpt(
        'insecure',
        default=False,
        help='If True, SSL/TLS certificate verification is disabled'
    ),
    cfg.StrOpt(
        'memcached_server',
        default=None,
        help='Url of memcached server to use for caching'
    ),
    cfg.IntOpt(
        'token_cache_time',
        default=60,
        min=0,
        help='In order to prevent excessive effort spent validating '
             'tokens, the middleware caches previously-seen tokens '
             'for a configurable duration (in seconds).'
    ),
]

CONF = cfg.CONF
CONF.register_opts(keycloak_oidc_opts, group="keycloak_oidc")


class KeycloakAuthMiddleware(base_middleware.Middleware):
    def __init__(self, app):
        super(KeycloakAuthMiddleware, self).__init__(application=app)
        mcserv_url = CONF.keycloak_oidc.memcached_server
        self.mcclient = memcache.Client(mcserv_url) if mcserv_url else None

        self.certfile = CONF.keycloak_oidc.certfile
        self.keyfile = CONF.keycloak_oidc.keyfile
        self.cafile = CONF.keycloak_oidc.cafile or utils.get_system_ca_file()
        self.insecure = CONF.keycloak_oidc.insecure
        self.url_template = CONF.keycloak_oidc.auth_url + \
            CONF.keycloak_oidc.user_info_endpoint_url

    def authenticate(self, access_token, realm_name):
        info = None
        if self.mcclient:
            info = self.mcclient.get(access_token)

        if info is None and CONF.keycloak_oidc.user_info_endpoint_url:

            url = self.url_template % realm_name

            verify = None
            if urllib.parse.urlparse(url).scheme == "https":
                verify = False if self.insecure else self.cafile

            cert = (self.certfile, self.keyfile) \
                if self.certfile and self.keyfile else None

            try:
                resp = requests.get(
                    url,
                    headers={"Authorization": "Bearer %s" % access_token},
                    verify=verify,
                    cert=cert
                )
            except requests.ConnectionError:
                msg = _("Can't connect to keycloak server with address '%s'."
                        ) % CONF.keycloak_oidc.auth_url
                LOG.error(msg)
                raise exception.GlareException(message=msg)

            if resp.status_code == 401:
                raise exception.Unauthorized(message=resp.text)
            if resp.status_code == 403:
                raise exception.Forbidden(message=resp.text)
            elif resp.status_code >= 400:
                raise exception.GlareException(message=resp.text)

            if self.mcclient:
                self.mcclient.set(access_token, resp.json(),
                                  time=CONF.keycloak_oidc.token_cache_time)
            info = resp.json()

        LOG.debug("HTTP response from OIDC provider: %s",
                  pprint.pformat(info))

        return info

    @webob.dec.wsgify
    def __call__(self, request):
        if 'X-Auth-Token' not in request.headers:
            msg = _("Auth token must be provided in 'X-Auth-Token' header.")
            LOG.error(msg)
            raise exception.Unauthorized()
        access_token = request.headers.get('X-Auth-Token')
        try:
            decoded = jwt.decode(access_token, algorithms=['RS256'],
                                 verify=False)
        except Exception:
            msg = _("Token can't be decoded because of wrong format.")
            LOG.error(msg)
            raise exception.Unauthorized()

        # Get user realm from parsed token
        # Format is "iss": "http://<host>:<port>/auth/realms/<realm_name>",
        __, __, realm_name = decoded['iss'].strip().rpartition('/realms/')

        # Get roles from from parsed token
        roles = ','.join(decoded['realm_access']['roles']) \
            if 'realm_access' in decoded else ''

        self.authenticate(access_token, realm_name)

        request.headers["X-Identity-Status"] = "Confirmed"
        request.headers["X-Project-Id"] = realm_name
        request.headers["X-Roles"] = roles
        return request.get_response(self.application)



# Copyright 2016 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_serialization import jsonutils
import webob

from glare.api import versions
from glare.tests.unit import base


class VersionsTest(base.BaseTestCase):

    """Test the version information returned from the API service."""

    def test_root_endpoint(self):
        req = webob.Request.blank('/', base_url='http://127.0.0.1:9494/')
        req.accept = 'application/json'
        res = versions.Controller().index(req)
        self.assertEqual(300, res.status_int)
        self.assertEqual('application/json', res.content_type)
        results = jsonutils.loads(res.body)['versions']
        expected = [
            {'links': [{'href': 'http://docs.openstack.org/',
                        'rel': 'describedby',
                        'type': 'text/html'}],
             'media-type': 'application/vnd.openstack.artifacts-1.0',
             'status': 'STABLE',
             'version': '1.0'},
            {'links': [{'href': 'http://docs.openstack.org/',
                        'rel': 'describedby',
                        'type': 'text/html'}],
             'media-type': 'application/vnd.openstack.artifacts-1.1',
             'status': 'EXPERIMENTAL',
             'version': '1.1'}]
        self.assertEqual(expected, results)

    def test_versions_endpoint(self):
        req = webob.Request.blank('/versions',
                                  base_url='http://127.0.0.1:9494/')
        req.accept = 'application/json'
        res = versions.Controller().index(req)
        self.assertEqual(300, res.status_int)
        self.assertEqual('application/json', res.content_type)
        results = jsonutils.loads(res.body)['versions']
        expected = [
            {'links': [{'href': 'http://docs.openstack.org/',
                        'rel': 'describedby',
                        'type': 'text/html'}],
             'media-type': 'application/vnd.openstack.artifacts-1.0',
             'status': 'STABLE',
             'version': '1.0'},
            {'links': [{'href': 'http://docs.openstack.org/',
                        'rel': 'describedby',
                        'type': 'text/html'}],
             'media-type': 'application/vnd.openstack.artifacts-1.1',
             'status': 'EXPERIMENTAL',
             'version': '1.1'}]
        self.assertEqual(expected, results)



# Copyright 2011 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
A filter middleware that inspects the requested URI for a version string
and/or Accept headers and attempts to negotiate an API controller to
return.
"""

import microversion_parse
from oslo_log import log as logging
from oslo_middleware import base as base_middleware


from glare.api.v1 import api_version_request as api_version
from glare.api import versions as artifacts_versions
from glare.common import exception

LOG = logging.getLogger(__name__)


class GlareVersionNegotiationFilter(base_middleware.ConfigurableMiddleware):
    """Middleware that defines API version in request and redirects it
    to correct Router.
    """

    SERVICE_TYPE = 'artifact'
    MIME_TYPE = 'application/vnd.openstack.artifacts-'

    @staticmethod
    def get_version_from_accept(accept_header):
        """Try to parse accept header to extract api version.

        :param accept_header: accept header
        :return: version string in the request or None if not specified
        """
        accept = str(accept_header)
        if accept.startswith(GlareVersionNegotiationFilter.MIME_TYPE):
            LOG.debug("Using media-type versioning")
            return accept[len(GlareVersionNegotiationFilter.MIME_TYPE):]

        return None

    @staticmethod
    def process_request(req):
        """Process api request:
        1. Define if this is request for available versions or not
        2. If it is not version request check extract version
        3. Validate available version and add version info to request
        """

        args = {'method': req.method, 'path': req.path, 'accept': req.accept}
        LOG.debug("Determining version of request: %(method)s %(path)s "
                  "Accept: %(accept)s", args)

        # determine if this is request for versions
        if req.path_info in ('/versions', '/'):
            return artifacts_versions.Controller.index(req)

        # determine api version from request
        req_version = GlareVersionNegotiationFilter.get_version_from_accept(
            req.accept)
        if req_version is None:
            # determine api version from microversion header
            LOG.debug("Determine version from microversion header.")
            req_version = microversion_parse.get_version(
                req.headers,
                service_type=GlareVersionNegotiationFilter.SERVICE_TYPE)

        # validate microversions header
        req.api_version_request = \
            GlareVersionNegotiationFilter._get_api_version_request(
                req_version)
        req_version = req.api_version_request.get_string()

        LOG.debug("Matched version: %s", req_version)
        LOG.debug('new path %s', req.path_info)

    @staticmethod
    def _get_api_version_request(req_version):
        """Set API version for request based on the version header string."""
        if req_version is None:
            LOG.debug("No API version in request header. Use default version.")
            cur_ver = api_version.APIVersionRequest.default_version()
        elif req_version == 'latest':
            # 'latest' is a special keyword which is equivalent to
            # requesting the maximum version of the API supported
            cur_ver = api_version.APIVersionRequest.max_version()
        else:
            cur_ver = api_version.APIVersionRequest(req_version)

        # Check that the version requested is within the global
        # minimum/maximum of supported API versions
        if not cur_ver.matches(cur_ver.min_version(), cur_ver.max_version()):
            raise exception.InvalidGlobalAPIVersion(
                req_ver=cur_ver.get_string(),
                min_ver=cur_ver.min_version().get_string(),
                max_ver=cur_ver.max_version().get_string())
        return cur_ver

    @staticmethod
    def process_response(response, request=None):
        if hasattr(response, 'headers'):
            if hasattr(request, 'api_version_request'):
                api_header_name = microversion_parse.STANDARD_HEADER
                response.headers[api_header_name] = (
                    GlareVersionNegotiationFilter.SERVICE_TYPE + ' ' +
                    request.api_version_request.get_string())
                response.headers.add('Vary', api_header_name)

        return response



# Copyright 2017 - Nokia Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.


import eventlet
from oslo_config import cfg
from oslo_log import log as logging

from glare.api.middleware import context
from glare.common import exception
from glare.common import store_api
from glare.db.sqlalchemy import api as db_api
from glare.i18n import _

LOG = logging.getLogger(__name__)

scrubber_opts = [
    cfg.IntOpt('scrub_time', default=0, min=0,
               help=_("""
The amount of time, in seconds, to delay artifact scrubbing.
When delayed delete is turned on, an artifact is put into
``deleted`` state upon deletion until the scrubber deletes its data.
Typically, soon
after the artifact is put into ``deleted`` state, it is available
for scrubbing. However, scrubbing can be delayed until a later point
using this configuration option. This option denotes the time period
an artifact spends in ``deleted`` state before it is available for
scrubbing.
It is important to realize that this has storage implications. The
larger the ``scrub_time``, the longer the time to reclaim backend
storage from deleted artifacts.
Possible values:
    * Any non-negative integer
Related options:
    * ``delayed_delete``
""")),
    cfg.IntOpt('scrub_pool_size', default=1, min=1,
               help=_("""
The size of thread pool to be used for scrubbing artifacts.
When there are a large number of artifacts to scrub, it is
beneficial to scrub artifacts in parallel so that the scrub queue
stays in control and the backend storage is reclaimed in a timely
fashion. This configuration option denotes the maximum number of
artifacts to be scrubbed in parallel. The default value is one,
which signifies serial scrubbing. Any value above one indicates
parallel scrubbing.
Possible values:
    * Any non-zero positive integer
Related options:
    * ``delayed_delete``
""")),
]

scrubber_cmd_opts = [
    cfg.IntOpt('wakeup_time', default=300, min=0,
               help=_("""
Time interval, in seconds, between scrubber runs in daemon mode.
Scrubber can be run either as a cron job or daemon. When run as a
daemon, this configuration time specifies the time period between
two runs. When the scrubber wakes up, it fetches and scrubs all
``deleted`` artifacts that are available for scrubbing after taking
``scrub_time`` into consideration.
If the ``wakeup_time`` is set to a large number, there may be a large
number of artifacts to be scrubbed for each run. Also, this impacts
how quickly the backend storage is reclaimed.
Possible values:
    * Any non-negative integer
Related options:
    * ``daemon``
    * ``delayed_delete``
"""))
]

scrubber_cmd_cli_opts = [
    cfg.BoolOpt('daemon',
                short='D',
                default=False,
                help=_("""
Run scrubber as a daemon.
This boolean configuration option indicates whether scrubber should
run as a long-running process that wakes up at regular intervals to
scrub artifacts. The wake up interval can be specified using the
configuration option ``wakeup_time``.
If this configuration option is set to ``False``, which is the
default value, scrubber runs once to scrub artifacts and exits.
In this case, if the operator wishes to implement continuous
scrubbing of artifacts, scrubber needs to be scheduled as a cron job.
Possible values:
    * True
    * False
Related options:
    * ``wakeup_time``
"""))
]

CONF = cfg.CONF
CONF.register_opts(scrubber_opts, group='scrubber')


class Daemon(object):
    def __init__(self, wakeup_time=300, threads=100):
        LOG.info("Starting Daemon: wakeup_time=%(wakeup_time)s "
                 "threads=%(threads)s",
                 {'wakeup_time': wakeup_time, 'threads': threads})
        self.wakeup_time = wakeup_time
        self.event = eventlet.event.Event()
        # This pool is used for periodic instantiation of scrubber
        self.daemon_pool = eventlet.greenpool.GreenPool(threads)

    def start(self, application):
        self._run(application)

    def wait(self):
        try:
            self.event.wait()
        except KeyboardInterrupt:
            LOG.info("Daemon Shutdown on KeyboardInterrupt")

    def _run(self, application):
        LOG.debug("Running scrubber application")
        self.daemon_pool.spawn_n(application.run, self.event)
        eventlet.spawn_after(self.wakeup_time, self._run, application)
        LOG.debug("Next run scheduled in %s seconds", self.wakeup_time)


class Scrubber(object):
    def __init__(self):
        self.context = context.RequestContext()
        self.context.is_admin = True
        self.pool = eventlet.greenpool.GreenPool(
            CONF.scrubber.scrub_pool_size)

    def run(self, event=None):
        while True:
            artifacts = db_api._get_all(
                context=self.context,
                session=db_api.get_session(),
                limit=CONF.scrubber.scrub_pool_size,
                sort=[],
                filters=[('status', None, 'eq', None, 'deleted')])
            if not artifacts:
                break
            self.pool.imap(self._scrub_artifact, artifacts)

    @staticmethod
    def _scrub_artifact(af):
        LOG.info("Begin scrubbing of artifact %s", af.id)
        for blob in af.blobs:
            if not blob.external:
                try:
                    store_api.delete_blob(blob.url, context=context)
                except exception.NotFound:
                    # data has already been removed
                    pass
        LOG.info("Blobs successfully deleted for artifact %s", af.id)

        # delete artifact itself
        db_api.delete(context, af.id, db_api.get_session())

        LOG.info("Artifact %s was scrubbed", af.id)
